import importlib.metadata

from .. import __package_name__


class StageNotFoundError(Exception):
    """Custom error for when a stage is not found."""


class StageDiscoveryError(Exception):
    """Custom error for stage discovery."""


def get_all_registered_stages() -> dict[str, importlib.metadata.EntryPoint]:
    """Returns a dictionary mapping stage name to entry point for all registered MEDS-transforms stages.

    Registration here _does not mean simply having used the `Stage.register` decorator_. Rather, it means
    having added a `MEDS_transform.stages` entry point to your python package. E.g., in your `pyproject.toml`,
    you could have:

    ```toml
    [project.entry-points."MEDS_transforms.stages"]
    aggregate_code_metadata = "MEDS_transforms.stages:aggregate_code_metadata"
    fit_vocabulary_indices = "MEDS_transforms.stages:fit_vocabulary_indices"
    ```

    If so, then this function will include the stage names `aggregate_code_metadata` and
    `fit_vocabulary_indices` in the returned dictionary, mapping to the entry point objects that you specify.

    For the examples below, we will merely mock the entry points so we can showcase behaviors, but in practice
    this would be read dynamically across all installed packages in the python context.

    Examples:
        >>> fake_eps = {
        ...     "wrong_package.stages": [("A", "stage_a"), ("B", "stage_b")], # [(name, entry_point), ...]
        ...     "MEDS_transforms.stages": [("C", "stage_c"), ("D", "stage_d")],
        ... }
        >>> def get_group(group: str):
        ...     vals = fake_eps[group]
        ...     out = MagicMock()
        ...     out.names = [name for name, _ in vals]
        ...     out.select = lambda name: [entry_point for n, entry_point in vals if n == name]
        ...     out.__getitem__ = lambda self, name: next(ep for n, ep in vals if n == name)
        ...     return out
        >>> with patch("importlib.metadata.entry_points") as mock_entry_points:
        ...     mock_entry_points.side_effect = get_group
        ...     get_all_registered_stages()
        {'C': 'stage_c', 'D': 'stage_d'}

    If there are duplicate stage names, we get an error:

        >>> fake_eps["MEDS_transforms.stages"].append(("C", "stage_c_2"))
        >>> with patch("importlib.metadata.entry_points") as mock_entry_points:
        ...     mock_entry_points.side_effect = get_group
        ...     get_all_registered_stages()
        Traceback (most recent call last):
            ...
        MEDS_transforms.stages.discovery.StageDiscoveryError: Multiple entry points registered for the same
        stage:
          - C: ['stage_c', 'stage_c_2']
    """
    eps = importlib.metadata.entry_points(group=f"{__package_name__}.stages")

    errors = {}
    for name in eps.names:
        if len(eps.select(name=name)) > 1 and name not in errors:
            errors[name] = eps.select(name=name)
    if errors:
        error_str = "\n".join(f"  - {name}: {eps}" for name, eps in errors.items())
        raise StageDiscoveryError(f"Multiple entry points registered for the same stage:\n{error_str}")

    return {n: eps[n] for n in eps.names}



"""This file defines two useful decorators for use in config management.

1. A decorator to define structured configs in Hydra's config store.
2. A decorator to annotate functions as OmegaConf resolvers in the global namespace.
"""

import dataclasses
from collections.abc import Callable
from functools import partial

from hydra.core.config_store import ConfigStore
from omegaconf import OmegaConf


def OmegaConfResolver(*args, **kwargs) -> Callable:  # noqa: N802
    """A decorator to register the decorated function as an OmegaConf resolver.

    If a name is passed via keyword arguments, it will be used as the name of the resolver. If no name is
    passed, the name of the function will be used. Other keyword arguments will be passed to the
    OmegaConf.register_new_resolver function.

    Args:
        args: The function to register as a resolver (if specified, no keyword arguments should be used)
        kwargs: The name of the resolver and other keyword arguments. If specified, this is a parametrized
            decorator and the function to register should not be explicitly passed. See the examples below.

    Returns:
        A decorator that registers the decorated function as an OmegaConf resolver.

    Raises:
        TypeError: If the arguments are not valid for the OmegaConfResolver decorator.

    Examples:
        >>> @OmegaConfResolver
        ... def my_resolver(x: int) -> str:
        ...     return str(x)

    After decorating, the function is registered as an OmegaConf resolver via its name, and can be used as a
    resolver in OmegaConf, but remains unchanged in the python scope.

        >>> OmegaConf.has_resolver("my_resolver")
        True
        >>> cfg = OmegaConf.create("foo: ${my_resolver:42}")
        >>> cfg.foo
        '42'
        >>> my_resolver(42)  # The function is unchanged
        '42'

    This can also be used as a parametrized decorator:

        >>> @OmegaConfResolver(name="my_resolver2")
        ... def my_resolver(x: int) -> str:
        ...     return str(x**2)
        >>> OmegaConf.has_resolver("my_resolver2")
        True
        >>> cfg = OmegaConf.create("foo: ${my_resolver2:8}")
        >>> cfg.foo
        '64'
        >>> my_resolver(8)  # The function is unchanged
        '64'

    Errors are raised if the arguments are not valid:

        >>> OmegaConfResolver("foo")
        Traceback (most recent call last):
            ...
        TypeError: Invalid arguments for OmegaConfResolver decorator. Expected either a positional function or
        keyword arguments; got ('foo',) and {}.
    """

    def decorator(func: Callable, name: str | None = None, **kwargs) -> Callable:
        if name is None:
            name = func.__name__
        OmegaConf.register_new_resolver(name, func, **kwargs)
        return func

    if len(args) == 0:
        return partial(decorator, **kwargs)
    elif len(args) == 1 and callable(args[0]) and not kwargs:
        return decorator(args[0])
    else:
        raise TypeError(
            "Invalid arguments for OmegaConfResolver decorator. Expected either a positional function or "
            f"keyword arguments; got {args} and {kwargs}."
        )


def hydra_registered_dataclass(*, group: str | None, name: str | None = None) -> Callable[[type], type]:
    """A simple decorator to define a class as a dataclass and register it with Hydra.

    This is a parametrized decorator that takes in the hydra config-store group (mandatory, must be specified
    by keyword) and name (optional, can be inferred from the decorated class). The flagged class will be
    registered in the hydra config store with the given name and made a dataclass.

    Args:
        group: The hydra config-store group to register the class with.
        name: The name to register the class with. If None, the class name will be used.

    Returns:
        A decorator that takes a class, makes it a dataclass, and registers it with Hydra.

    Examples:
        >>> @hydra_registered_dataclass(group="foo")
        ... class Bar:
        ...     baz: str = "baz"
        >>> dataclasses.is_dataclass(Bar)
        True
        >>> cs = ConfigStore.instance()
        >>> cs.repo["foo"]
        {'Bar.yaml': ConfigNode(name='Bar.yaml', node={'baz': 'baz'}, group='foo', ...)}
        >>> @hydra_registered_dataclass(group="foo", name="Bar2")
        ... class Bar:
        ...     qux: str = "quux"
        >>> cs.repo["foo"]
        {'Bar.yaml': ..., 'Bar2.yaml': ConfigNode(name='Bar2.yaml', node={'qux': 'quux'}, ...)}
    """

    def decorator(cls: type, name: str | None = None) -> type:
        if name is None:
            name = cls.__name__
        cls = dataclasses.dataclass(cls)
        cs = ConfigStore.instance()
        cs.store(group=group, name=name, node=cls)
        return cls

    return partial(decorator, name=name)



"""This file defines an object-oriented backing for configuring pipelines in MEDS-Transforms."""

from __future__ import annotations

import copy
import dataclasses
import logging
import os
from importlib.resources import files
from pathlib import Path
from typing import Any

from hydra.core.config_store import ConfigStore
from omegaconf import DictConfig, ListConfig, OmegaConf

from ..stages.base import Stage
from ..stages.discovery import StageNotFoundError, get_all_registered_stages
from .stage import UNPARSED_STAGE_T, StageConfig

logger = logging.getLogger(__name__)

NULL_STR = "__null__"
PKG_PFX = "pkg://"
YAML_EXTENSIONS = {".yaml", ".yml"}


def resolve_pkg_path(pkg_path: str) -> Path:
    """Parse a package path into a package name and a relative path.

    Args:
        pkg_path (str): The package path to parse.

    Returns:
        The file-path on disk to the package resource.

    Raises:
        ValueError: If the package path is not valid.

    Examples:
        >>> resolve_pkg_path("pkg://MEDS_transforms.configs.pipeline.py")
        PosixPath('...MEDS_transforms/configs/pipeline.py')

        Files need not exist to be returned:

        >>> resolve_pkg_path("pkg://MEDS_transforms.configs.pipeline.zip")
        PosixPath('...MEDS_transforms/configs/pipeline.zip')

        Note that this _returns something likely wrong_ for multi-suffix or no-suffix files!

        >>> resolve_pkg_path("pkg://MEDS_transforms.configs.pipeline")  # likely should end in /pipeline
        PosixPath('...MEDS_transforms/configs.pipeline')
        >>> resolve_pkg_path("pkg://MEDS_transforms.configs.data.tar.gz")  # likely should end in /data.tar.gz
        PosixPath('...MEDS_transforms/configs/data/tar.gz')

        Errors occur if the package is not importable:

        >>> resolve_pkg_path("pkg://non_existent_package.configs.pipeline.py")
        Traceback (most recent call last):
            ...
        ValueError: Package 'non_existent_package' not found. Please check the package name.
    """
    parts = pkg_path[len(PKG_PFX) :].split(".")
    pkg_name = parts[0]

    suffix = parts[-1]
    relative_path = Path(os.path.join(*parts[1:-1])).with_suffix(f".{suffix}")
    try:
        return files(pkg_name) / relative_path
    except ModuleNotFoundError as e:
        raise ValueError(f"Package '{pkg_name}' not found. Please check the package name.") from e


@dataclasses.dataclass
class PipelineConfig:
    """A base configuration class for MEDS-transforms pipelines.

    This class is used to define the structure of a pipeline configuration file. It manually tracks the
    necessary parameters (`stages` and `stage_configs`) and stores all other parameters in an
    `additional_params` `DictConfig`.

    It's primary use is to abstract functionality for resolving stage specific parameters to form the stage
    configuration object for a stage and pipeline realization from arguments.

    Attributes:
        stages: A list of (raw) stage configuration specifications that are part of the pipeline.
        additional_params: A dictionary of additional parameters that are not stage-specific.

    Raises:
        TypeError: If the stages or additional_params are not of the expected type.
        ValueError: If the pipeline configuration is invalid or if there are duplicate stage names.

    Examples:
        >>> PipelineConfig()
        PipelineConfig(stages=None, additional_params={})
        >>> stages = [{"stage1": {"param1": 1}}, "stage2"]
        >>> additional_params = {"param2": 2}
        >>> PipelineConfig(stages=stages, additional_params=additional_params)
        PipelineConfig(stages=[{'stage1': {'param1': 1}}, 'stage2'],
                       additional_params={'param2': 2})

    Some validation of the pipeline configuration is performed immediately after construction:

        >>> PipelineConfig(stages="invalid_type")
        Traceback (most recent call last):
            ...
        TypeError: Invalid type for stages: <class 'str'>. Expected list or ListConfig.
        >>> PipelineConfig(stages=[{3: {"param1": 1}}, "stage2"])
        Traceback (most recent call last):
            ...
        ValueError: Failed to parse pipeline configuration. Please check the pipeline YAML file 'stages' key.
        >>> PipelineConfig(stages=[{"stage1": {"param1": 1}}, "stage1"])
        Traceback (most recent call last):
            ...
        ValueError: Duplicate stage name found: stage1
        >>> PipelineConfig(stages=[{"stage1": {"param1": 1}}, "stage2"], additional_params=3)
        Traceback (most recent call last):
            ...
        TypeError: Invalid type for additional_params: <class 'int'>. Expected dict or DictConfig.

    While you can create a PipelineConfig object directly, it is more often created via the `from_arg` method,
    which loads a pipeline configuration from a YAML file. This YAML file is not structured precisely as the
    object is, as the additional parameters are flattened in the file representation. See the `from_arg`
    method for more details.

        >>> file_representation = {"stages": stages, **additional_params}
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as pipeline_yaml:
        ...     OmegaConf.save(file_representation, pipeline_yaml.name)
        ...     cfg = PipelineConfig.from_arg(pipeline_yaml.name)
        >>> cfg
        PipelineConfig(stages=[{'stage1': {'param1': 1}}, 'stage2'],
                       additional_params={'param2': 2})

    Once defined, you can also access the `parsed_stages` property to get a list of the stages in the pipeline
    parsed as `StageConfig` objects. This is how the user API for simple format of stages is translated to a
    readable format.

        >>> cfg.parsed_stages
        [StageConfig(name='stage1', base_stage=None, config={'param1': 1}),
         StageConfig(name='stage2', base_stage=None, config={})]
        >>> PipelineConfig().parsed_stages
        []
        >>> PipelineConfig(stages=[{"stage1": {"_base_stage": "foo"}}]).parsed_stages
        [StageConfig(name='stage1', base_stage='foo', config={})]

    Pipeline configurations can be resolved to a `DictConfig` representation suitable for Hydra registration
    via the `structured_config` property. This `DictConfig` representation has `additional_params` flattened
    into the output, much like the file representation. This representation will omit missing or empty keys
    automatically as well:

        >>> cfg.structured_config
        {'stages': [{'stage1': {'param1': 1}}, 'stage2'], 'param2': 2}
        >>> PipelineConfig().structured_config
        {}

    We can also use the `PipelineConfig` for a pipeline to prepare to run a stage in that pipeline, by
    resolving the relevant input and output file paths given the prior pipeline stages, resolving stages to
    their associated base (runnable) stage names, and registering the necessary structured configuration nodes
    within the Hydra `ConfigStore` for runtime use. These behaviors all take place via the `register_for`
    method. This method both (a) identifies and loads the runnable stage specified by the argument (and
    returns it), and (b) constructs the Hydra nodes and adds them to the `ConfigStore`. Let's first inspect
    the returned stage:

        >>> pipeline_cfg = PipelineConfig(stages=["filter_subjects"])
        >>> runnable_stage = pipeline_cfg.register_for("filter_subjects")
        >>> print(runnable_stage)
        Stage filter_subjects:
          Type: map
          is_metadata: False
          Docstring:
            | Returns a function that filters subjects by the number of measurements and events they have.
            ...
          Default config:
            | min_events_per_subject: null
            | min_measurements_per_subject: null
          Map function: filter_subjects
          Reduce function: None
          Main function: None
          Mimic function: filter_subjects

    And to see the registered configuration node, we can use the `ConfigStore` to get the node:

        >>> cs = ConfigStore.instance()
        >>> cs.repo["_pipeline.yaml"].node
        {'stages': ['filter_subjects'],
         'stage_cfg': {'min_events_per_subject': None, 'min_measurements_per_subject': None,
                       'data_input_dir': '${input_dir}/data', 'metadata_input_dir': '${input_dir}/metadata',
                       'reducer_output_dir': None, 'train_only': False, 'output_dir': '${output_dir}/data'}}

    The simplest pipeline, a "null" pipeline, with `PipelineConfig()` or
    `PipelineConfig.from_arg("__null__")`, will not have any stages or stage configurations. When you call
    `register_for` on such a pipeline, it will automatically add the stage being registered to the pipeline;
    this reflects that "null" pipelines are created automatically when running singleton stages.

        >>> pipeline_cfg = PipelineConfig()
        >>> runnable_stage = pipeline_cfg.register_for("occlude_outliers")
        >>> pipeline_cfg.stages
        ['occlude_outliers']

    If the pipeline indicates that a stage actually is an instance of a base stage, that will be reflected in
    the resolution. Similarly, arguments specified in the stage configs will appear in the resolved
    `stage_cfg` node. In addition, `register_for` automatically takes into account other stages in the
    pipeline:

        >>> pipeline_cfg = PipelineConfig(
        ...     stages=[
        ...         {
        ...             "count_codes": {"aggregations": ["code/n_subjects", "code/n_occurrences"]},
        ...             "_base_stage": "aggregate_code_metadata",
        ...         },
        ...         {"filter_measurements": {"min_subjects_per_code": 4, "min_occurrences_per_code": 10}},
        ...         {
        ...             "fit_outlier_detection": {
        ...                 "aggregations": ["values/n_occurrences", "values/sum", "values/sum_sqd"],
        ...             },
        ...             "_base_stage": "aggregate_code_metadata",
        ...         },
        ...         "occlude_outliers",
        ...     ],
        ... )
        >>> runnable_stage = pipeline_cfg.register_for("count_codes")
        >>> print(runnable_stage.stage_name)
        aggregate_code_metadata
        >>> cs.repo["_pipeline.yaml"].node
        {'stages': [{'count_codes': {'aggregations': ['code/n_subjects', 'code/n_occurrences']},
                     '_base_stage': 'aggregate_code_metadata'},
                    {'filter_measurements': {'min_subjects_per_code': 4, 'min_occurrences_per_code': 10}},
                    {'fit_outlier_detection': {'aggregations': ['values/n_occurrences',
                                                                'values/sum',
                                                                'values/sum_sqd']},
                     '_base_stage': 'aggregate_code_metadata'},
                    'occlude_outliers'],
         'stage_cfg': {'data_input_dir': '${input_dir}/data',
                       'metadata_input_dir': '${input_dir}/metadata',
                       'output_dir': '${output_dir}/count_codes',
                       'train_only': True,
                       'reducer_output_dir': '${output_dir}/count_codes',
                       'aggregations': ['code/n_subjects', 'code/n_occurrences']}}
        >>> runnable_stage = pipeline_cfg.register_for("filter_measurements")
        >>> print(runnable_stage.stage_name)
        filter_measurements
        >>> cs.repo["_pipeline.yaml"].node
        {'stages': [{'count_codes': {'aggregations': ['code/n_subjects', 'code/n_occurrences']},
                     '_base_stage': 'aggregate_code_metadata'},
                    {'filter_measurements': {'min_subjects_per_code': 4, 'min_occurrences_per_code': 10}},
                    {'fit_outlier_detection': {'aggregations': ['values/n_occurrences',
                                                                'values/sum',
                                                                'values/sum_sqd']},
                     '_base_stage': 'aggregate_code_metadata'},
                    'occlude_outliers'],
         'stage_cfg': {'min_subjects_per_code': 4,
                       'min_occurrences_per_code': 10,
                       'data_input_dir': '${input_dir}/data',
                       'metadata_input_dir': '${output_dir}/count_codes',
                       'reducer_output_dir': None,
                       'train_only': False,
                       'output_dir': '${output_dir}/filter_measurements'}}

    If you try to register a pipeline that includes an unregistered stage, an error will be raised:

        >>> pipeline_cfg = PipelineConfig(stages=["foobar"])
        >>> pipeline_cfg.register_for("foobar")
        Traceback (most recent call last):
            ...
        ValueError: Stage 'foobar' not registered! Registered stages: ...
    """

    stages: list[UNPARSED_STAGE_T] | ListConfig | None = None
    additional_params: dict[str, Any] | DictConfig = dataclasses.field(default_factory=dict)

    @classmethod
    def from_arg(cls, arg: str | Path) -> PipelineConfig:
        """Construct a pipeline configuration object from a specified pipeline YAML file.

        Args:
            arg: The path to the pipeline YAML file on disk or in the
                'pkg://<pkg_name>.<relative_path>' format. It can also be the sentinel `__null__` string,
                which will return an empty PipelineConfig object.

        Returns:
            A PipelineConfig object corresponding to the specified pipeline YAML file. Note that this object
            will not exactly match the passed file; rather, the stages and stage configurations will be pulled
            out separately and stored as direct attributes, and the rest of the parameters will be stored in
            the `additional_params` attribute.

        Raises:
            TypeError: If the pipeline YAML path is not a string or Path object.
            ValueError: If the pipeline YAML path does not have a valid file extension.
            FileNotFoundError: If the pipeline YAML file does not exist.

        Examples:
            >>> PipelineConfig.from_arg("__null__")
            PipelineConfig(stages=None, additional_params={})
            >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as pipeline_yaml:
            ...     OmegaConf.save({"stages": ["stage1", "stage2"], "foobar": 3}, pipeline_yaml.name)
            ...     PipelineConfig.from_arg(pipeline_yaml.name)
            PipelineConfig(stages=['stage1', 'stage2'], additional_params={'foobar': 3})

            To show the package path resolution, we can use the `pkg://` format, but for this test, we need to
            mock the package structure with unittest.mock.patch:

            >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as pipeline_yaml:
            ...     pipeline_fp = Path(pipeline_yaml.name)
            ...     OmegaConf.save({"stages": ["stage1", "stage2"], "qux": "a"}, pipeline_fp)
            ...     with patch("MEDS_transforms.configs.pipeline.resolve_pkg_path", return_value=pipeline_fp):
            ...         PipelineConfig.from_arg("pkg://fake_pkg.pipeline.yaml")
            PipelineConfig(stages=['stage1', 'stage2'], additional_params={'qux': 'a'})

            It will throw errors if the file has the wrong extension, does not exist, or is a directory, and
            if the passed parameter is neither a string nor a Path object:

            >>> PipelineConfig.from_arg(3)
            Traceback (most recent call last):
                ...
            TypeError: Invalid pipeline YAML path type <class 'int'>. Expected str or Path.
            >>> with tempfile.NamedTemporaryFile(suffix=".json") as pipeline_yaml:
            ...     OmegaConf.save({"stages": ["stage1", "stage2"], "foobar": 3}, pipeline_yaml.name)
            ...     PipelineConfig.from_arg(pipeline_yaml.name)
            Traceback (most recent call last):
                ...
            ValueError: Invalid pipeline YAML path '/tmp/tmp....json'. Expected a file with one of the
                following extensions: ['.yaml', '.yml']
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            ...     pipeline_yaml = Path(tmpdir) / "pipeline.yaml"
            ...     PipelineConfig.from_arg(pipeline_yaml)
            Traceback (most recent call last):
                ...
            FileNotFoundError: Pipeline YAML file '/tmp/tmp.../pipeline.yaml' does not exist.
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            ...     PipelineConfig.from_arg(tmpdir)
            Traceback (most recent call last):
                ...
            FileNotFoundError: Pipeline YAML file '/tmp/tmp...' is a directory, not a file!
        """

        match arg:
            case str() if arg == NULL_STR:
                return cls()
            case str() as pkg_path if arg.startswith(PKG_PFX):
                pipeline_fp = resolve_pkg_path(pkg_path)
            case str() | Path() as path:
                pipeline_fp = Path(path)
            case _:
                raise TypeError(f"Invalid pipeline YAML path type {type(arg)}. Expected str or Path.")

        if pipeline_fp.exists() and pipeline_fp.is_dir():
            raise FileNotFoundError(f"Pipeline YAML file '{pipeline_fp}' is a directory, not a file!")
        elif pipeline_fp.suffix not in YAML_EXTENSIONS:
            raise ValueError(
                f"Invalid pipeline YAML path '{pipeline_fp}'. "
                f"Expected a file with one of the following extensions: {sorted(YAML_EXTENSIONS)}"
            )
        elif not pipeline_fp.exists():
            raise FileNotFoundError(f"Pipeline YAML file '{pipeline_fp}' does not exist.")

        as_dict_config = OmegaConf.load(pipeline_fp)

        stages = as_dict_config.pop("stages", None)
        return cls(stages=stages, additional_params=as_dict_config)

    def __post_init__(self):
        if self.stages is not None and not isinstance(self.stages, list | ListConfig):
            raise TypeError(f"Invalid type for stages: {type(self.stages)}. Expected list or ListConfig.")

        try:
            self.parsed_stages  # noqa: B018
        except Exception as e:
            raise ValueError(
                "Failed to parse pipeline configuration. Please check the pipeline YAML file 'stages' key."
            ) from e

        duplicate_stages = set()
        for s in self.parsed_stages:
            if s.name in duplicate_stages:
                raise ValueError(f"Duplicate stage name found: {s.name}")
            duplicate_stages.add(s.name)

        if not isinstance(self.additional_params, DictConfig | dict):
            raise TypeError(
                f"Invalid type for additional_params: {type(self.additional_params)}. "
                f"Expected dict or DictConfig."
            )

    @property
    def structured_config(self) -> DictConfig:
        """Return a `DictConfig` representation of the pipeline configuration suitable for Hydra registration.

        This `DictConfig` representation has `additional_params` flattened into the main configuration. This
        means that were you to store the `DictConfig` in a file, then re-load it as a `PipelineConfig` via
        `PipelineConfig.from_arg`, it would match the original `PipelineConfig` object.
        """

        merged = {}
        if self.stages is not None:
            merged["stages"] = self.stages
        if self.additional_params:
            merged.update(self.additional_params)
        return OmegaConf.create(merged)

    @property
    def parsed_stages(self) -> list[StageConfig]:
        """Return a list of `StageConfig` objects representing the stages in the pipeline.

        This property parses the `stages` attribute and returns a list of `StageConfig` objects, which
        represent the stages in the pipeline. The `StageConfig` objects are created by parsing the
        `stages` attribute and resolving any base stage names.
        """
        if self.stages is None:
            return []

        return [StageConfig.from_arg(s) for s in copy.deepcopy(self.stages)]

    @property
    def parsed_stages_by_name(self) -> dict[str, StageConfig]:
        return {s.name: s for s in self.parsed_stages}

    def _resolve_stages(self, all_stages: dict[str, Stage]) -> dict[str, DictConfig]:
        stage_objects = []
        last_data_stage_name = None
        last_metadata_stage_name = None
        for s in self.parsed_stages:
            stage = all_stages[s.resolved_name]

            if stage.is_metadata:
                last_metadata_stage_name = s.name
            else:
                last_data_stage_name = s.name
            stage_objects.append((s.name, stage, {**s.config}))

        prior_data_stage = None
        prior_metadata_stage = None

        resolved_stage_configs = {}

        input_dir = Path("${input_dir}")
        output_dir = Path("${output_dir}")

        for name, stage, config_overwrites in stage_objects:
            config = {**stage.default_config} if stage.default_config else {}

            if prior_data_stage is None:
                config["data_input_dir"] = str(input_dir / "data")
            else:
                config["data_input_dir"] = prior_data_stage["output_dir"]

            if prior_metadata_stage is None:
                config["metadata_input_dir"] = str(input_dir / "metadata")
            else:
                config["metadata_input_dir"] = prior_metadata_stage["reducer_output_dir"]

            if stage.is_metadata:
                config["output_dir"] = str(output_dir / name)
                config["train_only"] = True
                if name == last_metadata_stage_name:
                    config["reducer_output_dir"] = str(output_dir / "metadata")
                else:
                    config["reducer_output_dir"] = str(output_dir / name)
            else:
                config["reducer_output_dir"] = None
                config["train_only"] = False
                if name == last_data_stage_name:
                    config["output_dir"] = str(output_dir / "data")
                else:
                    config["output_dir"] = str(output_dir / name)

            config.update({k: v for k, v in config_overwrites.items() if k != "_base_stage"})
            resolved_stage_configs[name] = OmegaConf.create(config)

            if stage.is_metadata:
                prior_metadata_stage = config
            else:
                prior_data_stage = config

        return resolved_stage_configs

    def _resolve_stage_name(self, stage_name: str) -> str:
        """Return the registered stage corresponding to the specified stage for the given pipeline.

        Args:
            stage_name: The name of the stage to resolve.

        Returns: Either (a) the `_base_stage` specified in the pipeline config's `stage_configs` for this
            stage, if specified, or (b) the stage name itself, otherwise. In both cases, the stage name
            returned is validated to be a registered stage.

        Raises:
            ValueError: If the stage name is not in the pipeline configuration.
            StageNotFoundError: If the stage name is not a registered stage.

        Examples:
            >>> PipelineConfig(stages=["occlude_outliers"])._resolve_stage_name("occlude_outliers")
            'occlude_outliers'
            >>> PipelineConfig(stages=["occlude_outliers"])._resolve_stage_name("foobar")
            Traceback (most recent call last):
                ...
            ValueError: Stage foobar not in pipeline configuration!
            >>> PipelineConfig(stages=["foobar"])._resolve_stage_name("foobar")
            Traceback (most recent call last):
                ...
            MEDS_transforms.stages.discovery.StageNotFoundError: Stage 'foobar' not
                registered! Registered stages: ...

        Stage names can be resolved to a base stage name, if specified in the stage configs:

            >>> cfg = PipelineConfig(stages=[{"count_codes": {"_base_stage": "aggregate_code_metadata"}}])
            >>> cfg._resolve_stage_name("count_codes")
            'aggregate_code_metadata'
        """

        if stage_name not in self.parsed_stages_by_name:
            raise ValueError(f"Stage {stage_name} not in pipeline configuration!")

        resolved_stage_name = self.parsed_stages_by_name[stage_name].resolved_name

        all_stages = get_all_registered_stages()
        if resolved_stage_name not in all_stages:
            raise StageNotFoundError(
                f"Stage '{resolved_stage_name}' not registered! Registered stages: "
                f"{', '.join(sorted(all_stages.keys()))}"
            )

        return resolved_stage_name

    @Stage.suppress_validation()
    def register_for(self, stage_name: str) -> Stage:
        if not self.stages:
            logger.warning("No stages specified in the pipeline config. Adding the target stage alone.")
            self.stages = [stage_name]

        registered_stages = get_all_registered_stages()
        loaded_stages = {}
        for s in self.parsed_stages:
            if s.resolved_name not in registered_stages:
                raise ValueError(
                    f"Stage '{s.resolved_name}' not registered! Registered stages: "
                    f"{', '.join(sorted(registered_stages.keys()))}"
                )
            if s.resolved_name not in loaded_stages:
                loaded_stages[s.resolved_name] = registered_stages[s.resolved_name].load()

        resolved_stage_name = self._resolve_stage_name(stage_name)

        pipeline_node = self.structured_config
        pipeline_node["stage_cfg"] = self._resolve_stages(loaded_stages)[stage_name]

        cs = ConfigStore.instance()
        cs.store(name="_pipeline", node=pipeline_node)

        return loaded_stages[resolved_stage_name]



"""This module defines test helpers for defining derived packages integrating with MEDS-Transforms.

Pytest Plugin Capabilities:
    - CLI options for specifying the packages and stages within those packages to include in automated testing
      fixtures.
    - The `stage_example` fixture which exposes a stage and test scenario for using built-in stage testing
      support.
    - A helper (`pipeline_tester`) to test a full pipeline by specifying a list of stage scenarios that should
      capture the pipeline's behavior when run in sequence.
    - A helper context manager (`suppress_logging`) to suppress logging. This is useful if you include in your
      package some pytest setup code that iterates through partially defined stage examples, as many warnings
      may be logged by MEDS testing helpers upon partial initialization of dataset examples.
"""

import logging
import subprocess
import tempfile
import tomllib
from collections.abc import Callable
from contextlib import contextmanager
from pathlib import Path

import pytest
from omegaconf import OmegaConf

from . import __package_name__
from .configs.stage import StageConfig
from .stages import StageExample, get_all_registered_stages

logger = logging.getLogger(__name__)

# Get all registered stages


@contextmanager
def suppress_logging(level: int = logging.CRITICAL):
    """Suppress logging at the specified level within the context block.

    Args:
        level: The logging level at or below which logs will be suppressed. Defaults to CRITICAL.

    Examples:
        >>> import logging
        >>> logger = logging.getLogger(__name__)
        >>> with print_warnings():
        ...     logger.warning("This will be logged, as suppress logging isn't used.")
        Warning: This will be logged, as suppress logging isn't used.
        >>> with print_warnings(), suppress_logging():
        ...     logger.warning("This will not be logged.")
        >>> with print_warnings():
        ...     logger.warning("This will be logged again, as the suppression has been disabled.")
        Warning: This will be logged again, as the suppression has been disabled.

    Note that if you (for some reason) use both `print_warnings` and `suppress_logging` (as we do here so we
    can show that `suppress_logging` works -- in normal usage you'd never need to do this), the order of the
    two matters; if you reverse what we showed above, things will be printed:

        >>> with suppress_logging(), print_warnings():
        ...     logger.warning("Wrong order! This will be logged.")
        Warning: Wrong order! This will be logged.
    """
    logger = logging.getLogger()
    original_level = logger.getEffectiveLevel()
    logger.setLevel(level)
    try:
        yield
    finally:
        logger.setLevel(original_level)


def pytest_addoption(parser):  # pragma: no cover
    stages_for_package_help_str = (
        "Specify the package name(s) to search within when looking for registered stages to test. "
        "If not specified, all registered stages, _except for the core MEDS-Transforms stages_ will be "
        "tested. This option will attempt to automatically apply an intelligent default to just the current "
        "package being tested by pytest, but this may fail, in which case you can set this manually. "
        "This option can be used multiple times to specify multiple packages; e.g., "
        "'--test_stage_for_package package1 --test_stage_for_package package2'. "
        "Packages specified here but that lack any stages or any stages with static data examples will be "
        "skipped without error."
    )
    parser.addoption(
        "--test_stages_for_package", action="append", type=str, default=[], help=stages_for_package_help_str
    )

    stage_help_str = (
        "Specify the registered MEDS-Transforms stage(s) to test. If not specified, all registered "
        "stages will be run. This option can be used multiple times to specify multiple stages; e.g., "
        "'--test_stage stage1 --test_stage stage2'. The stages are specified by their registered names. "
        "This option only affects the automated testing of all registered stages, and does not impact other "
        "tests, such as the runner or any manual tests. Stages specified here but that lack static data "
        "examples will be skipped."
    )
    parser.addoption("--test_stage", action="append", type=str, default=[], help=stage_help_str)


def pytest_configure(config: pytest.Config):
    """Configure pytest by attaching the appropriate package to use to the config object.

    This adds the following attributes to the config object:
        - `allowed_stages`: A dictionary of the registered stages that are allowed to be run.
        - `allowed_stage_scenarios`: A dictionary of the registered stages and their scenarios that are
          set-up for automated testing.

    It does so based on the command line options passed to pytest (to determine the packages and stages to set
    up for testing). If no command line stages are passed, all stages in the allowed packages will be
    included. If no command line packages are passed, all stages in the current package will be included, if
    the current package can be auto-detected; otherwise all packages _except_ for the base MEDS-Transforms
    package will be tested (as this is only likely to happen in downstream derived packages, and you don't
    want to re-test all the base stages in such a case).

    This function also checks for any invalid stages specified on the command line and raises an error if
    any are found.

    Args:
        config: The pytest config object. It will be modified by having the above attributes added to it.

    Raises:
        ValueError: If any invalid stages are specified on the command line.

    Examples:
        >>> mock_stages = {
        ...     "stage_1_1": MagicMock(),
        ...     "stage_1_2": MagicMock(),
        ...     "stage_2_1": MagicMock(),
        ...     "stage_meds_transforms": MagicMock(),
        ... }
        >>> mock_stages["stage_1_1"].dist.metadata.__getitem__.return_value = "package1"
        >>> mock_stages["stage_1_2"].dist.metadata.__getitem__.return_value = "package1"
        >>> mock_stages["stage_2_1"].dist.metadata.__getitem__.return_value = "package2"
        >>> mock_stages["stage_meds_transforms"].dist.metadata.__getitem__.return_value = __package_name__
        >>> mock_stages["stage_1_1"].load.return_value.test_cases = ["test_case1", "test_case2"]
        >>> mock_stages["stage_1_2"].load.return_value.test_cases = ["test_case3", "test_case4"]
        >>> mock_stages["stage_2_1"].load.return_value.test_cases = ["test_case5", "test_case6"]
        >>> mock_stages["stage_meds_transforms"].load.return_value.test_cases = ["test_case7", "test_case8"]

    Let's also set up a fake config with CLI options

        >>> CLI_options = {
        ...     "test_stages_for_package": ["package1", "package2"],
        ...     "test_stage": ["stage_1_1", "stage_2_1"],
        ... }
        >>> config = MagicMock()
        >>> config.getoption.side_effect = lambda x: CLI_options.get(x, [])

    Now, if we call the function with the fake config and stages, we should get the expected output. This
    simulates reading the command line options and setting up the config object.

        >>> with patch("MEDS_transforms.pytest_plugin.get_all_registered_stages", return_value=mock_stages):
        ...     pytest_configure(config)
        >>> list(config.allowed_stages.keys())
        ['stage_1_1', 'stage_2_1']
        >>> config.allowed_stage_scenarios
        {'stage_1_1': ['test_case1', 'test_case2'], 'stage_2_1': ['test_case5', 'test_case6']}

    What if we change the config options?

        >>> CLI_options["test_stages_for_package"] = ["package1"]
        >>> CLI_options["test_stage"] = []
        >>> with patch("MEDS_transforms.pytest_plugin.get_all_registered_stages", return_value=mock_stages):
        ...     pytest_configure(config)
        >>> list(config.allowed_stages.keys())
        ['stage_1_1', 'stage_1_2']
        >>> config.allowed_stage_scenarios
        {'stage_1_1': ['test_case1', 'test_case2'], 'stage_1_2': ['test_case3', 'test_case4']}

    What if we request a stage not in the package?

        >>> CLI_options["test_stage"] = ["stage_1_1", "stage_2_1", "invalid_stage"]
        >>> with patch("MEDS_transforms.pytest_plugin.get_all_registered_stages", return_value=mock_stages):
        ...     pytest_configure(config)
        Traceback (most recent call last):
            ...
        ValueError: Invalid stage(s) specified for ['package1']: invalid_stage, stage_2_1.

    If we don't pass any CLI options, then it should try to auto detect the package. But, if that doesn't work
    (which we'll simulate here), it should test everything except MEDS-Transforms.

        >>> CLI_options["test_stages_for_package"] = []
        >>> CLI_options["test_stage"] = []
        >>> with patch("MEDS_transforms.pytest_plugin.get_all_registered_stages", return_value=mock_stages):
        ...     with patch("MEDS_transforms.pytest_plugin._auto_detect_package", return_value=None):
        ...         pytest_configure(config)
        >>> list(config.allowed_stages.keys())
        ['stage_1_1', 'stage_1_2', 'stage_2_1']
        >>> config.allowed_stage_scenarios
        {'stage_1_1': ['test_case1', 'test_case2'],
         'stage_1_2': ['test_case3', 'test_case4'],
         'stage_2_1': ['test_case5', 'test_case6']}
    """

    if cli_packages := config.getoption("test_stages_for_package"):
        packages_to_test = cli_packages
    elif detected_package := _auto_detect_package(Path(config.rootpath)):
        packages_to_test = [detected_package]
    else:
        packages_to_test = []

    stages = config.getoption("test_stage")

    registered_stages = get_all_registered_stages()

    allowed_stages = {}
    for n, ep in registered_stages.items():
        pkg = ep.dist.metadata["Name"]

        no_packages_specified = len(packages_to_test) == 0
        pkg_valid = (no_packages_specified and (pkg != __package_name__)) or (pkg in packages_to_test)

        stage_valid = (not stages) or (n in stages)

        if pkg_valid and stage_valid:
            allowed_stages[n] = ep.load()

    missing_stages = sorted(set(stages) - set(allowed_stages.keys()))

    if missing_stages:
        raise ValueError(f"Invalid stage(s) specified for {packages_to_test}: {', '.join(missing_stages)}.")

    config.allowed_stages = allowed_stages
    with suppress_logging():
        config.allowed_stage_scenarios = {n: s.test_cases for n, s in allowed_stages.items()}


def _auto_detect_package(root: Path) -> str | None:
    """Attempts to automatically detect the package name based on the pytest Config's rootpath.

    This is a best-effort attempt to find the package name by looking for a pyproject.toml file in the
    rootpath of the pytest config (passed directly). If this file is not found, or if the package name cannot
    be determined, `None` is returned.

    > [!Warning]
    > This function _only_ works with `pyproject.toml` files that are stored in the `root` path. `setup.py`
    > files are not supported!

    Args:
        root: The rootpath of the pytest session.

    Returns:
        The package name if found, otherwise `None`.

    Examples:
        >>> toml_contents = '''
        ... [project]
        ... name = "my_package"
        ... '''
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     pyproject_fp = Path(tmpdir) / "pyproject.toml"
        ...     _ = pyproject_fp.write_text(toml_contents)
        ...     print(_auto_detect_package(Path(tmpdir)))
        my_package

    If we can't find a pyproject.toml file or if it can't be parsed, we return `None`.

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     print(_auto_detect_package(Path(tmpdir)))  # No toml file!
        None
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     pyproject_fp = Path(tmpdir) / "pyproject.toml"
        ...     _ = pyproject_fp.write_text("Invalid TOML")
        ...     print(_auto_detect_package(Path(tmpdir)))
        None

    If a setup.py file is found, we warn the user that it is not supported and return `None`.

        >>> setup_contents = '''
        ... from setuptools import setup
        ... setup(
        ...     name="my_package",
        ...     version="0.1",
        ...     packages=["my_package"],
        ... )
        ... '''
        >>> with tempfile.TemporaryDirectory() as tmpdir, print_warnings():
        ...     setup_fp = Path(tmpdir) / "setup.py"
        ...     _ = setup_fp.write_text(setup_contents)
        ...     print(_auto_detect_package(Path(tmpdir)))
        None
        Warning: setup.py file found in /tmp/tmp..., but no pyproject.toml. Auto detection of package
        name only works with pyproject.toml files currently, so `None` will be returned.
    """

    pyproject_file = root / "pyproject.toml"
    setup_file = root / "setup.py"

    if pyproject_file.exists():
        try:
            with open(pyproject_file, "rb") as f:
                pyproject = tomllib.load(f)
                return pyproject.get("project", {}).get("name")
        except Exception:  # pragma: no cover
            pass
    elif setup_file.exists():
        logger.warning(
            f"setup.py file found in {root}, but no pyproject.toml. Auto detection of package name only "
            "works with pyproject.toml files currently, so `None` will be returned."
        )
    return None


def pytest_generate_tests(metafunc):
    """Generate tests for registered stages based on the command line options."""
    config = metafunc.config

    if "stage_scenario" in metafunc.fixturenames:
        arg_names = ["stage", "stage_scenario"]
        arg_values = [(s, sc) for s, scenarios in config.allowed_stage_scenarios.items() for sc in scenarios]
        metafunc.parametrize(arg_names, arg_values, scope="session")

    elif "stage" in metafunc.fixturenames:
        metafunc.parametrize("stage", list(config.allowed_stage_scenarios.keys()), scope="session")


@pytest.fixture(scope="session")
def stage_example(request: pytest.FixtureRequest, stage: str, stage_scenario: str) -> StageExample:
    """Fixture to provide the example for the given stage and scenario."""

    if stage_scenario not in request.config.allowed_stage_scenarios[stage]:  # pragma: no cover
        raise ValueError(f"Stage scenario '{stage_scenario}' not found for stage '{stage}'.")

    s = request.config.allowed_stages[stage]
    if s.stage_name != stage:  # pragma: no cover
        raise ValueError(
            f"Stage '{stage}' has a misconfigured registration point!. It is registered at {stage}, "
            f"but the loaded stage name is {s.name}."
        )

    yield request.config.allowed_stage_scenarios[stage][stage_scenario]


def pipeline_tester(
    pipeline_yaml: str,
    stage_runner_yaml: str | None,
    stage_scenario_sequence: list[str],
    run_fn: Callable | None = subprocess.run,
):
    """Test the pipeline with the given YAML configuration and stage scenario sequence.

    Args:
        pipeline_yaml: A string with the yaml for the pipeline configuration to be tested. If you would like
            to support a file or pkg file syntax here, file a GitHub issue.
        stage_runner_yaml: A string with the yaml for the stage runner configuration to be tested. If you
            would like to support a file or pkg file syntax here, file a GitHub issue.
        stage_scenario_sequence: A list of strings of the form `"{stage_name}/{scenario}"` where `"scenario"`
            may be "/" separated or non-existent (so just "{stage_name}") that define the input/output
            relationships expected for the pipeline for each stage -- e.g., the input to the first listed
            scenario is the test input to the pipeline, its output is expected to be the output that the
            pipeline produces when it runs the first stage, which is then fed into the second stage, etc.
            These are used to test the pipeline conforms to expectations across all stages.
        run_fn: A callable to run the pipeline. Defaults to `subprocess.run`. This is useful for dependency
            injection during testing.

    Raises:
        ValueError: If the pipeline YAML and stage scenario sequence do not match in length.
        AssertionError: If the pipeline fails to produce expected output for any stage.

    Examples:
        >>> pipeline_yaml = "stages: ['foo', 'bar']"
        >>> pipeline_tester(pipeline_yaml, "", ["just_one"])
        Traceback (most recent call last):
            ...
        ValueError: Incorrect pipeline test specification! Pipeline YAML has 2 stages, but 1 stage scenarios
        were provided.
        >>> pipeline_tester(pipeline_yaml, "", ["foo/just_one", "bar/just_one"])
        Traceback (most recent call last):
            ...
        ValueError: Error loading stage example for ...

    To see it in action, we'll use fake run functions.

        >>> def fake_run_success(script, shell, capture_output):
        ...     return subprocess.CompletedProcess([], returncode=0, stdout=b"Success", stderr=b"")
        >>> def fake_run_failure(script, shell, capture_output):
        ...     return subprocess.CompletedProcess([], returncode=1, stdout=b"", stderr=b"Failure")

    We'll use the real, default example for the `filter_subjects` stage here.

        >>> pipeline_yaml = "stages: [filter_subjects]"  # This wouldn't work in real life

    If we run and throw an error from the fake shell runner, it will raise an AssertionError.

        >>> pipeline_tester(pipeline_yaml, "", ["filter_subjects"], run_fn=fake_run_failure)
        Traceback (most recent call last):
            ...
        AssertionError: Pipeline failed with error: Pipeline returned code 1.
        Stdout:
        <BLANKLINE>
        Stderr:
        Failure

    If we run and succeed, it will still throw an error as the stage will fail to validate its expected files.

        >>> pipeline_tester(pipeline_yaml, "", ["filter_subjects"], run_fn=fake_run_success)
        Traceback (most recent call last):
            ...
        AssertionError: Pipeline failed to produce expected output for stage 'filter_subjects'
    """

    pipeline_stages = [StageConfig.from_arg(s).name for s in OmegaConf.create(pipeline_yaml).stages]

    if len(pipeline_stages) != len(stage_scenario_sequence):
        raise ValueError(
            "Incorrect pipeline test specification! "
            f"Pipeline YAML has {len(pipeline_stages)} stages, but {len(stage_scenario_sequence)} "
            "stage scenarios were provided."
        )

    registered_stages = get_all_registered_stages()

    stage_examples = []
    for stage_scenario in stage_scenario_sequence:
        parts = stage_scenario.split("/")
        if len(parts) > 1:
            stage_name, scenario_name = parts[0], "/".join(parts[1:])
        else:
            stage_name = stage_scenario
            scenario_name = "."

        try:
            stage_examples.append(registered_stages[stage_name].load().test_cases[scenario_name])
        except Exception as e:
            raise ValueError(f"Error loading stage example for {stage_name}/{scenario_name}") from e

    with tempfile.TemporaryDirectory() as tmpdir:
        # 1. Set-up
        test_root = Path(tmpdir)

        input_dir = test_root / "input"
        input_dir.mkdir(parents=True, exist_ok=True)

        stage_examples[0].write_for_test(input_dir)

        output_dir = test_root / "output"

        pipeline_yaml = pipeline_yaml.format(input_dir=input_dir, output_dir=output_dir)

        pipeline_config_fp = test_root / "pipeline.yaml"
        pipeline_config_fp.write_text(pipeline_yaml)

        command = [
            "MEDS_transform-pipeline",
            f"pipeline_config_fp={pipeline_config_fp}",
        ]

        if stage_runner_yaml is not None:
            stage_runner_fp = test_root / "stage_runner.yaml"
            stage_runner_fp.write_text(stage_runner_yaml)
            command.append(f"stage_runner_fp={stage_runner_fp}")

        # 2. Run the pipeline
        out = run_fn(
            command,
            shell=False,
            capture_output=True,
        )

        def err_msg(m: str) -> str:
            lines = [
                f"Pipeline failed with error: {m}",
                "Stdout:",
                out.stdout.decode("utf-8"),
                "Stderr:",
                out.stderr.decode("utf-8"),
            ]
            return "\n".join(lines)

        assert out.returncode == 0, err_msg(f"Pipeline returned code {out.returncode}.")

        # 3. Check the output
        last_data_stage = (None, None)
        last_metadata_stage = (None, None)

        for name, stage in zip(pipeline_stages, stage_examples, strict=True):
            if stage.want_data is not None:
                last_data_stage = (name, stage)
            if stage.want_metadata is not None:
                last_metadata_stage = (name, stage)

        for name, stage in zip(pipeline_stages, stage_examples, strict=True):
            is_last_data_stage = last_data_stage[0] == name
            is_last_metadata_stage = last_metadata_stage[0] == name

            stage_output_dir = output_dir / name

            stage.df_check_kwargs = {"rtol": 5e-2, "atol": 1e-2, "check_dtypes": False}

            try:
                if stage.want_data is not None:
                    if is_last_data_stage:
                        stage.check_outputs(output_dir)
                    else:
                        stage.check_outputs(stage_output_dir, is_resolved_dir=True)

                if stage.want_metadata is not None:
                    if is_last_metadata_stage:
                        stage.check_outputs(output_dir)
                    else:
                        stage.check_outputs(stage_output_dir, is_resolved_dir=True)
            except AssertionError as e:
                raise AssertionError(f"Pipeline failed to produce expected output for stage '{name}'") from e



"""Basic utilities for serialized reduce operations on sharded MEDS datasets with caching and locking."""

import logging
import time
from pathlib import Path
from typing import Protocol

import polars.selectors as cs

from ..dataframe import DF_T, READ_FN_T, WRITE_FN_T

logger = logging.getLogger(__name__)


class REDUCE_FN_T(Protocol):  # noqa: N801
    """Protocol for a function that takes a variable number dataframes and returns one dataframe."""

    def __call__(self, *dfs: DF_T) -> DF_T: ...


def reduce_over(
    in_fps: list[Path],
    out_fp: Path,
    read_fn: READ_FN_T,
    write_fn: WRITE_FN_T,
    reduce_fn: REDUCE_FN_T,
    merge_fp: Path | None = None,
    merge_fn: REDUCE_FN_T | None = None,
    do_overwrite: bool = False,
    polling_time: float = 0.1,
):
    """Performs a reduction operation on a list of input file paths, with optional merging to existing data.

    Args:
        in_fps: List of input file paths containing data over which the reduction should be performed.
        out_fp: Output file path where the reduced data will be saved.
        polling_time: Time in seconds to wait between checks for file readiness.
        read_fn: Function to read data from the input file paths.
        write_fn: Function to write the reduced data to the output file path.
        reduce_fn: Function to perform the reduction operation on the data. It should take two dataframe
            arguments and return one dataframe
        merge_fp: If this file exists, merge the output with the data stored in this file before finalization.
        merge_fn: A special reducer to perform the merging, if `merge_fp` is specified and is a file.
        do_overwrite: Should this overwrite an existing out file?

    Raises:
        FileExistsError: If the output file already exists.

    Examples:
        >>> def reduce_fn(*dfs: pl.DataFrame) -> pl.DataFrame:
        ...     return pl.concat(dfs, how="vertical")
        >>> kwargs = {
        ...     "read_fn": pl.read_parquet,
        ...     "write_fn": pl.DataFrame.write_parquet,
        ...     "reduce_fn": reduce_fn,
        ... }
        >>> dfs = [pl.DataFrame({"a": [i, i + 1], "b": [i + 2, i + 3]}) for i in range(3)]
        >>> def write_dfs(in_fps: list[Path], dfs: list[pl.DataFrame] = dfs, delay_per: float = 0):
        ...     for i, (fp, df) in enumerate(zip(in_fps, dfs)):
        ...         time.sleep(delay_per)  # simulate slow write
        ...         df.write_parquet(fp)
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     in_fps = [Path(tmpdir) / f"input_{i}.parquet" for i in range(3)]
        ...     write_dfs(in_fps)
        ...     out_fp = Path(tmpdir) / "output.parquet"
        ...     reduce_over(in_fps, out_fp, **kwargs)
        ...     pl.read_parquet(out_fp)
        shape: (6, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i8  ┆ i8  │
        ╞═════╪═════╡
        │ 0   ┆ 2   │
        │ 1   ┆ 3   │
        │ 1   ┆ 3   │
        │ 2   ┆ 4   │
        │ 2   ┆ 4   │
        │ 3   ┆ 5   │
        └─────┴─────┘

    If specified, the reducer will merge with the `merge_fp` via the specialized function

        >>> merge_df = pl.DataFrame({"a": [-1, -2], "b": [-3, -4]})
        >>> def merge_fn(new: pl.DataFrame, old: pl.DataFrame) -> pl.DataFrame:
        ...     return pl.concat([old, new], how="vertical_relaxed")
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     in_fps = [Path(tmpdir) / f"input_{i}.parquet" for i in range(3)]
        ...     write_dfs(in_fps)
        ...     merge_fp = Path(tmpdir) / "merge.parquet"
        ...     merge_df.write_parquet(merge_fp)
        ...     out_fp = Path(tmpdir) / "output.parquet"
        ...     reduce_over(in_fps, out_fp, merge_fp=merge_fp, merge_fn=merge_fn, **kwargs)
        ...     pl.read_parquet(out_fp)
        shape: (8, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i64 ┆ i64 │
        ╞═════╪═════╡
        │ -1  ┆ -3  │
        │ -2  ┆ -4  │
        │ 0   ┆ 2   │
        │ 1   ┆ 3   │
        │ 1   ┆ 3   │
        │ 2   ┆ 4   │
        │ 2   ┆ 4   │
        │ 3   ┆ 5   │
        └─────┴─────┘


    The reducer will wait for all the input files to exist before performing the computation:

        >>> import threading
        >>> def profile_reduce(dfs: list[pl.DataFrame], in_fps: list[Path], delay_per: float, out_fp: Path):
        ...     print("------------------")
        ...     print(f"Writing files with a delay of {delay_per} seconds...")
        ...     thread = threading.Thread(target=write_dfs, args=(in_fps, dfs, delay_per))
        ...     thread.daemon = True
        ...     st = datetime.now()
        ...     thread.start()
        ...     print("Starting reduction...")
        ...     reduce_over(in_fps, out_fp, **kwargs)
        ...     print(f"Reduction completed in: ~{(datetime.now() - st).total_seconds():.1f} seconds")
        ...     print(pl.read_parquet(out_fp))
        ...     thread.join()
        ...     for fp in in_fps:
        ...         fp.unlink()
        ...     out_fp.unlink()
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     in_fps = [Path(tmpdir) / f"input_{i}.parquet" for i in range(3)]
        ...     out_fp = Path(tmpdir) / "output.parquet"
        ...     profile_reduce(dfs, in_fps, 0.1, out_fp)
        ...     profile_reduce(dfs, in_fps, 0.5, out_fp)
        ------------------
        Writing files with a delay of 0.1 seconds...
        Starting reduction...
        Reduction completed in: ~0.4 seconds
        shape: (6, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i8  ┆ i8  │
        ╞═════╪═════╡
        │ 0   ┆ 2   │
        │ 1   ┆ 3   │
        │ 1   ┆ 3   │
        │ 2   ┆ 4   │
        │ 2   ┆ 4   │
        │ 3   ┆ 5   │
        └─────┴─────┘
        ------------------
        Writing files with a delay of 0.5 seconds...
        Starting reduction...
        Reduction completed in: ~1.5 seconds
        shape: (6, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i8  ┆ i8  │
        ╞═════╪═════╡
        │ 0   ┆ 2   │
        │ 1   ┆ 3   │
        │ 1   ┆ 3   │
        │ 2   ┆ 4   │
        │ 2   ┆ 4   │
        │ 3   ┆ 5   │
        └─────┴─────┘

    The reducer will error if the output file already exists, unless `do_overwrite` is set to `True`:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     in_fps = [Path(tmpdir) / f"input_{i}.parquet" for i in range(3)]
        ...     write_dfs(in_fps)
        ...     out_fp = Path(tmpdir) / "output.parquet"
        ...     out_fp.touch()
        ...     reduce_over(in_fps, out_fp, **kwargs)
        Traceback (most recent call last):
            ...
        FileExistsError: Output file already exists: /tmp/tmp.../output.parquet
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     in_fps = [Path(tmpdir) / f"input_{i}.parquet" for i in range(3)]
        ...     write_dfs(in_fps)
        ...     out_fp = Path(tmpdir) / "output.parquet"
        ...     out_fp.touch()
        ...     reduce_over(in_fps, out_fp, do_overwrite=True, **kwargs)
        ...     pl.read_parquet(out_fp)
        shape: (6, 2)
        ┌─────┬─────┐
        │ a   ┆ b   │
        │ --- ┆ --- │
        │ i8  ┆ i8  │
        ╞═════╪═════╡
        │ 0   ┆ 2   │
        │ 1   ┆ 3   │
        │ 1   ┆ 3   │
        │ 2   ┆ 4   │
        │ 2   ┆ 4   │
        │ 3   ┆ 5   │
        └─────┴─────┘
    """

    if out_fp.is_file() and not do_overwrite:
        raise FileExistsError(f"Output file already exists: {out_fp.resolve()!s}")

    while not all(fp.is_file() for fp in in_fps):
        logger.info("Waiting to begin reduction for all files to be written...")
        time.sleep(polling_time)

    reduced = reduce_fn(*[read_fn(fp) for fp in in_fps]).with_columns(cs.numeric().shrink_dtype().name.keep())

    if merge_fp is not None and merge_fp.is_file():
        reduced = merge_fn(reduced, read_fn(merge_fp))

    out_fp.parent.mkdir(parents=True, exist_ok=True)
    write_fn(reduced, out_fp)



"""Utilities for re-sharding a MEDS cohort to subsharded splits."""

import json
import logging
import math
import time
from collections import defaultdict
from collections.abc import Sequence
from functools import partial
from pathlib import Path

import numpy as np
import polars as pl
from meds import DataSchema, SubjectSplitSchema, subject_splits_filepath
from omegaconf import DictConfig

from ...dataframe import read_and_filter_fntr, write_df
from ...mapreduce.rwlock import rwlock_wrap
from ...mapreduce.shard_iteration import shard_iterator, shuffle_shards
from .. import Stage

logger = logging.getLogger(__name__)


def shard_subjects(
    subjects: np.ndarray,
    n_subjects_per_shard: int = 50000,
    external_splits: dict[str, Sequence[int]] | None = None,
    split_fracs_dict: dict[str, float] | None = {"train": 0.8, "tuning": 0.1, "held_out": 0.1},  # noqa: B006
    seed: int = 1,
) -> dict[str, list[int]]:
    """Shard a list of subjects, nested within train/tuning/held-out splits.

    This function takes a list of subjects and shards them into train/tuning/held-out splits, with the shards
    of a consistent size, nested within the splits. The function will also respect external splits, if
    provided, such that mandated splits (such as prospective held out sets or pre-existing, task-specific held
    out sets) are with-held and sharded as separate splits from the IID splits defined by `split_fracs_dict`.
    It returns a dictionary mapping the split and shard names (realized as f"{split}/{shard}") to the list of
    subjects in that shard.

    Args:
        subjects: The list of subjects to shard.
        n_subjects_per_shard: The maximum number of subjects to include in each shard.
        external_splits: The externally defined splits to respect. If provided, the keys of this dictionary
            will be used as split names, and the values as the list of subjects in that split. These
            pre-defined splits will be excluded from IID splits generated by this function, but will be
            sharded like normal. Note that this is largely only appropriate for held-out sets for pre-defined
            tasks or test cases (e.g., prospective tests); training subjects should often still be included in
            the IID splits to maximize the amount of data that can be used for training.
        split_fracs_dict: A dictionary mapping the split name to the fraction of subjects to include in that
            split. Defaults to 80% train, 10% tuning, 10% held-out. This can be None or empty only when
            external splits fully specify the population.
        seed: The random seed to use for shuffling the subjects before seeding and sharding. This is useful
            for ensuring reproducibility.

    Returns:
        A dictionary mapping f"{split}/{shard}" to the list of subjects in that shard. This may include
        overlapping subjects across a subset of these splits, but never across shards within a split. Any
        overlap will solely occur between the an external split and another external split.

    Raises:
        ValueError: If the sum of the split fractions in `split_fracs_dict` is not equal to 1.

    Examples:
        >>> subjects = np.array([1, 2, 3, 4, 5, 6, 7, 8, 9, 10], dtype=int)
        >>> shard_subjects(subjects, n_subjects_per_shard=3)
        {'train/0': [9, 4, 8], 'train/1': [2, 1, 10], 'train/2': [6, 5], 'tuning/0': [3], 'held_out/0': [7]}
        >>> shard_subjects(subjects, 3, split_fracs_dict={"train": 0.8, "tuning": 0.2, "held_out": None})
        {'train/0': [5, 9, 6], 'train/1': [3, 10, 8], 'train/2': [1, 2], 'tuning/0': [7, 4]}
        >>> shard_subjects(subjects, 3, split_fracs_dict={"train": 0.8, "held_out": None})
        Traceback (most recent call last):
            ...
        ValueError: The sum of the split fractions must be equal to 1. Got 0.8 through {'train': 0.8}.
        >>> external_splits = {
        ...     "taskA/held_out": np.array([8, 9, 10], dtype=int),
        ...     "taskB/held_out": np.array([10, 8, 9], dtype=int),
        ... }
        >>> shard_subjects(subjects, 3, external_splits)
        {'train/0': [5, 7, 4],
         'train/1': [1, 2],
         'tuning/0': [3],
         'held_out/0': [6],
         'taskA/held_out/0': [8, 9, 10],
         'taskB/held_out/0': [10, 8, 9]}
        >>> shard_subjects(subjects, n_subjects_per_shard=3, split_fracs_dict={"train": 0.5})
        Traceback (most recent call last):
            ...
        ValueError: The sum of the split fractions must be equal to 1. Got 0.5 through {'train': 0.5}.
        >>> shard_subjects([1, 2], n_subjects_per_shard=3)
        Traceback (most recent call last):
            ...
        ValueError: Unable to adjust splits to ensure all splits have at least 1 subject.
        >>> external_splits = {
        ...     "train": np.array([1, 2, 3, 4, 5, 6], dtype=int),
        ...     "test": np.array([7, 8, 9, 10], dtype=int),
        ... }
        >>> shard_subjects(subjects, 6, external_splits, split_fracs_dict=None)
        {'train/0': [1, 2, 3, 4, 5, 6], 'test/0': [7, 8, 9, 10]}
        >>> shard_subjects(subjects, 3, external_splits)
        {'train/0': [5, 1, 3], 'train/1': [2, 6, 4], 'test/0': [10, 7], 'test/1': [8, 9]}
    """

    if external_splits is None:
        external_splits = {}
    else:
        for k in list(external_splits.keys()):
            if not isinstance(external_splits[k], np.ndarray):
                logger.warning(
                    f"External split {k} is not a numpy array and thus type safety is not guaranteed. "
                    f"Attempting to convert to numpy array of dtype {subjects.dtype}."
                )
                external_splits[k] = np.array(external_splits[k], dtype=subjects.dtype)

    subjects = np.unique(subjects)

    # Splitting
    all_external_splits = set().union(*external_splits.values())
    is_in_external_split = np.isin(subjects, list(all_external_splits))
    subject_ids_to_split = subjects[~is_in_external_split]

    splits = external_splits

    if split_fracs_dict is not None and None in split_fracs_dict.values():
        filtered_split_fracs_dict = {k: v for k, v in split_fracs_dict.items() if v is not None}
        logger.info(
            "Ignoring splits with null fraction: "
            + ", ".join(set(split_fracs_dict).difference(filtered_split_fracs_dict))
        )
        split_fracs_dict = filtered_split_fracs_dict

    splits_cover = sum(split_fracs_dict.values()) if split_fracs_dict else 0

    rng = np.random.default_rng(seed)
    if n_subjects := len(subject_ids_to_split):
        if not math.isclose(splits_cover, 1):
            raise ValueError(
                f"The sum of the split fractions must be equal to 1. Got {splits_cover} "
                f"through {split_fracs_dict}."
            )
        split_names_idx = rng.permutation(len(split_fracs_dict))
        split_names = np.array(list(split_fracs_dict.keys()))[split_names_idx]
        split_fracs = np.array([split_fracs_dict[k] for k in split_names])
        split_lens = np.round(split_fracs[:-1] * n_subjects).astype(int)
        split_lens = np.append(split_lens, n_subjects - split_lens.sum())

        if split_lens.min() == 0:
            logger.warning(
                "Some splits are empty. Adjusting splits to ensure all splits have at least 1 subject."
            )
            max_split = split_lens.argmax()
            split_lens[max_split] -= 1
            split_lens[split_lens.argmin()] += 1

        if split_lens.min() == 0:
            raise ValueError("Unable to adjust splits to ensure all splits have at least 1 subject.")

        subjects = rng.permutation(subject_ids_to_split)
        subjects_per_split = np.split(subjects, split_lens.cumsum())

        splits = {**dict(zip(split_names, subjects_per_split, strict=False)), **splits}
    else:
        if split_fracs_dict:
            logger.warning(
                "External splits were provided covering all subjects, but split_fracs_dict was not empty. "
                "Ignoring the split_fracs_dict."
            )
        else:
            logger.info("External splits were provided covering all subjects.")

    # Sharding
    final_shards = {}
    for sp, pts in splits.items():
        if len(pts) <= n_subjects_per_shard:
            final_shards[f"{sp}/0"] = pts.tolist()
        else:
            pts = rng.permutation(pts)
            n_pts = len(pts)
            n_shards = int(np.ceil(n_pts / n_subjects_per_shard))
            shards = np.array_split(pts, n_shards)
            for i, shard in enumerate(shards):
                final_shards[f"{sp}/{i}"] = shard.tolist()

    seen = {}

    for k, pts in final_shards.items():
        logger.info(f"Split {k} has {len(pts)} subjects.")

        for kk, v in seen.items():
            shared = set(pts).intersection(v)
            if shared:
                logger.info(f"  - intersects {kk} on {len(shared)} subjects.")

        seen[k] = set(pts)

    return final_shards


def valid_json_file(fp: Path) -> bool:
    """Check if a file is a valid JSON file.

    Args:
        fp: Path to the file.

    Returns:
        True if the file is a valid JSON file, False otherwise.

    Examples:
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     fp = Path(tmpdir) / "test.json"
        ...     valid_json_file(fp)
        False
        >>> with tempfile.NamedTemporaryFile(suffix=".json") as tmpfile:
        ...     fp = Path(tmpfile.name)
        ...     _ = fp.write_text("foobar not a json file.\tHello, world!")
        ...     valid_json_file(fp)
        False
        >>> with tempfile.NamedTemporaryFile(suffix=".json") as tmpfile:
        ...     fp = Path(tmpfile.name)
        ...     _ = fp.write_text('{"foo": "bar"}')
        ...     valid_json_file(fp)
        True
    """
    if not fp.is_file():
        return False
    try:
        json.loads(fp.read_text())
        return True
    except json.JSONDecodeError:
        return False


def make_new_shards_fn(df: pl.DataFrame, cfg: DictConfig, stage_cfg: DictConfig) -> dict[str, list[str]]:
    """This function creates a new sharding scheme for the MEDS cohort."""
    splits_map = defaultdict(list)
    for pt_id, sp in df.iter_rows():
        splits_map[sp].append(pt_id)

    return shard_subjects(
        subjects=df[SubjectSplitSchema.subject_id_name].to_numpy(),
        n_subjects_per_shard=stage_cfg.n_subjects_per_shard,
        external_splits=splits_map,
        split_fracs_dict=None,
        seed=cfg.get("seed", 1),
    )


def write_json(d: dict, fp: Path) -> None:
    """Write a dictionary to a JSON file.

    Args:
        d: Dictionary to write.
        fp: Path to the file.

    Examples:
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     fp = Path(tmpdir) / "test.json"
        ...     write_json({"foo": "bar"}, fp)
        ...     fp.read_text()
        '{"foo": "bar"}'
    """
    fp.write_text(json.dumps(d))


@Stage.register(is_metadata=False)
def main(cfg: DictConfig):
    """Re-shard a MEDS cohort to in a manner that subdivides subject splits."""

    output_dir = Path(cfg.stage_cfg.output_dir)

    splits_file = Path(cfg.input_dir) / subject_splits_filepath
    shards_fp = output_dir / ".shards.json"

    rwlock_wrap(
        splits_file,
        shards_fp,
        partial(pl.read_parquet, use_pyarrow=True),
        write_json,
        partial(make_new_shards_fn, cfg=cfg, stage_cfg=cfg.stage_cfg),
        do_overwrite=cfg.do_overwrite,
        out_fp_checker=valid_json_file,
    )

    max_iters = cfg.get("max_iters", 10)
    iters = 0
    while not valid_json_file(shards_fp) and iters < max_iters:  # pragma: no cover
        logger.info(f"Waiting to begin until shards map is written. Iteration {iters}/{max_iters}...")
        time.sleep(cfg.polling_time)
        iters += 1

    new_sharded_splits = json.loads(shards_fp.read_text())

    orig_shards_iter, _ = shard_iterator(cfg, out_suffix="")

    orig_shards_iter = [(in_fp, out_fp.relative_to(output_dir)) for in_fp, out_fp in orig_shards_iter]

    new_shards = shuffle_shards(list(new_sharded_splits.keys()), cfg)
    new_shards_iter = [(shard_name, output_dir / f"{shard_name}.parquet") for shard_name in new_shards]

    # Step 1: Sub-sharding stage
    logger.info("Starting sub-sharding")

    for subshard_name, out_fp in new_shards_iter:
        subjects = new_sharded_splits[subshard_name]

        read_df = read_and_filter_fntr(pl.col(DataSchema.subject_id_name).is_in(subjects))

        def read_fn(input_dir: Path) -> pl.LazyFrame:
            df = None
            logger.info(f"Reading shards for {subshard_name} (file names are in the input sharding scheme):")  # noqa: B023
            for in_fp, _ in orig_shards_iter:
                logger.info(f"  - {in_fp.relative_to(input_dir).resolve()!s}")
                if df is None:
                    df = read_df(in_fp)  # noqa: B023
                else:
                    df = df.merge_sorted(read_df(in_fp), key=DataSchema.subject_id_name)  # noqa: B023

            return df

        def compute_fn(df: list[pl.DataFrame]) -> pl.LazyFrame:
            sort_cols = [DataSchema.subject_id_name, DataSchema.time_name]
            return df.sort(by=sort_cols, maintain_order=True, multithreaded=False)

        logger.info(f"Merging sub-shards for {subshard_name} to {out_fp.resolve()!s}")
        rwlock_wrap(
            cfg.stage_cfg.data_input_dir,
            out_fp,
            read_fn,
            write_df,
            compute_fn,
            do_overwrite=cfg.do_overwrite,
        )

    logger.info(f"Done with {cfg.stage}")



"""Functions for registering and defining MEDS-transforms stages."""

from __future__ import annotations

import copy
import inspect
import logging
import os
import sys
import textwrap
import warnings
from collections.abc import Callable
from contextlib import contextmanager
from enum import StrEnum
from functools import partial, wraps
from pathlib import Path
from typing import TYPE_CHECKING, Any, ClassVar

from omegaconf import DictConfig, OmegaConf

from ..mapreduce import map_stage, mapreduce_stage
from .discovery import get_all_registered_stages
from .examples import StageExample, StageExampleDict

if TYPE_CHECKING:
    import polars as pl

    from ..compute_modes import ANY_COMPUTE_FN_T
    from ..dataframe import READ_FN_T, WRITE_FN_T

logger = logging.getLogger(__name__)

MAIN_FN_T = Callable[[DictConfig], None]
VALIDATION_ENV_VAR = "DISABLE_STAGE_VALIDATION"


class StageType(StrEnum):
    """The types of stages MEDS-Transforms supports.

    Attributes:
        MAP: A stage that applies a transformation to each data shard of a MEDS dataset, outputting new data
            shards.
        MAPREDUCE: A stage that applies a metadata extraction operation to each data shard of a MEDS dataset,
            then reduces the outputs of those transformations into an updated metadata/codes.parquet file.
        MAIN: A stage that does not fit into either of the above categories, and provides a direct main
            function.
    """

    MAP = "map"
    MAPREDUCE = "mapreduce"
    MAIN = "main"

    @classmethod
    def from_fns(
        cls, main_fn: MAIN_FN_T | None, map_fn: ANY_COMPUTE_FN_T | None, reduce_fn: ANY_COMPUTE_FN_T | None
    ) -> StageType:
        """Determines the stage type based on the provided functions.

        Args:
            main_fn: The main function for the stage. May be None.
            map_fn: The mapping function for the stage. May be None.
            reduce_fn: The reducing function for the stage. May be None.

        Returns:
            StageType: The type of stage determined from the provided functions. If the passed functions do
                not correspond to a valid stage type, a ValueError will be raised.

        Raises:
            ValueError: If the provided functions do not correspond to a valid stage type.

        Examples:
            >>> def main(cfg: DictConfig):
            ...     pass
            >>> def map_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     pass
            >>> def reduce_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     pass
            >>> StageType.from_fns(main_fn=main, map_fn=None, reduce_fn=None)
            <StageType.MAIN: 'main'>
            >>> StageType.from_fns(main_fn=None, map_fn=map_fn, reduce_fn=None)
            <StageType.MAP: 'map'>
            >>> StageType.from_fns(main_fn=None, map_fn=map_fn, reduce_fn=reduce_fn)
            <StageType.MAPREDUCE: 'mapreduce'>
            >>> StageType.from_fns(main_fn=None, map_fn=None, reduce_fn=None)
            Traceback (most recent call last):
                ...
            ValueError: Either main_fn or map_fn/reduce_fn must be provided.
            >>> StageType.from_fns(main_fn=main, map_fn=map_fn, reduce_fn=reduce_fn)
            Traceback (most recent call last):
                ...
            ValueError: Only one of main_fn or map_fn/reduce_fn should be provided.
        """

        if main_fn is not None:
            if map_fn is not None or reduce_fn is not None:
                raise ValueError("Only one of main_fn or map_fn/reduce_fn should be provided.")
            return StageType.MAIN
        elif map_fn is not None and reduce_fn is not None:
            return StageType.MAPREDUCE
        elif map_fn is not None:
            return StageType.MAP
        else:
            raise ValueError("Either main_fn or map_fn/reduce_fn must be provided.")


class StageRegistrationError(Exception):
    pass


class Stage:
    """The representation of a MEDS-Transforms stage, in object form.

    Largely speaking, this is just a container around the different component functions that are used to
    execute the stage, as well as stage specific metadata like the name, docstring, and type.

    When constructed through the decorator `Stage.register`, it will be set to mimic in usage the function
    being decorated, though will still be a Stage object. This is done through the use of the `__call__`
    operator, and ensures that things like doctests and imports of the function being decorated do not behave
    unexpectedly. When constructed through the `Stage` constructor directly or through non-decorator usage,
    instead, the `__call__` operator will point to the main function of the stage. This enables two styles of
    usage; one where a map or compute function is decorated, but if the resulting function is imported or used
    in doctests it will behave like the decorated function, and the main callable for the stage is only
    accessible in the `main` variable of the stage object, and the second where a stage is constructed
    manually and (typically) assigned to a `main` variable in a stage python script or equivalent.

    This class is usually created through the `Stage.register` method -- either as a decorator, or as a
    function directly called. See the examples below for more details.

    Calling `print` on a `Stage` object will display a simple method indicating the name of the stage, the
    global stage docstring, and the names of the main, map, reduce, and mimic functions, if set. This is
    largely for testing and debugging purposes, and is not intended to be used in production code.

    Static data examples can also be attached to the stage object, to enable automated testing and
    documentation of the stage. This is managed through the `examples_dir` attribute, which can be passed in
    upon construction. When constructed through the `Stage.register` function, which is the typical usage, if
    unspecified this will automatically be set in accordance with the following rules:
      1. If the location in which `Stage.register` is called lives in a directory of the same name as the
         stage being defined _and_ has a subdirectory called `examples`, then that directory will be used.
      2. Otherwise, it will remain unset and no examples will be attached.
    If you would like additional modes of default inference to be added (such as on the basis of the locations
    in which the passed functions are defined) please file a GitHub Issue.

    Attributes:
        stage_type: The type of stage this is. This is set automatically based on the provided functions. See
            `StageType` for more details.
        stage_name: The name of the stage. This is set automatically based on the provided functions or can be
            set manually.
        stage_docstring: The docstring of the stage. This is set automatically based on the provided functions
            or can be set manually.
        map_fn: The mapping function for the stage. This is set automatically based on the provided functions.
            May be None for a `StageType.MAIN` stage.
        reduce_fn: The reducing function for the stage. This is set automatically based on the provided
            functions. May be None for a `StageType.MAP` or `StageType.MAIN` stage.
        main_fn: The main function for the stage. This is set automatically based on the provided functions.
        examples_dir: A directory containing nested test cases for the stage.
            If not set, this is automatically inferred in the case that the stage name and registering file
            conform to the pattern mentioned above.
        output_schema_updates: A dictionary mapping column name to a Polars type for the output of the stage,
            with the base MEDS schema options as defaults for unspecified columns.
        default_config: A dictionary containing the default configuration options for the stage. This can be
            passed manually during registration or is set automatically based on the calling file location in
            a manner similar to the examples directory.
        is_metadata: A boolean indicating whether the stage is a metadata stage. This, in some cases, is set
            automatically based on the functions passed or can be manually overridden.

    Examples:
        >>> def compute(cfg: DictConfig):
        ...     '''base fn docstring'''
        ...     return "compute"
        >>> stage = Stage(map_fn=compute)

    At its base, a stage is a container for a function that can be used to run the stage. Most of the examples
    shown here are focused on more internal aspects of how stages work; for documentation on how you will most
    likely use stages, see the documentation for the `Stage.register` function.

    Stages come with tracked test cases and default configuration arguments. If no special parameters are set
    during construction, they will be empty:

        >>> print(stage.examples_dir)
        None
        >>> stage.test_cases
        {}
        >>> print(stage.default_config)
        {}

    This is because test cases are inferred through the `examples_dir` attribute. This can be set manually:

        >>> stage = Stage(map_fn=compute, examples_dir=Path("foo"))
        >>> print(stage.examples_dir)
        foo

    But the test cases will only exist if the directory is properly set up:

        >>> stage.test_cases
        {}

    If the example dir is not a `Path` and is not `None`, an error will be raised:

        >>> stage = Stage(map_fn=compute, examples_dir="foo")
        Traceback (most recent call last):
            ...
        TypeError: examples_dir must be a Path or None. Got <class 'str'>: foo

    Likewise, the default config can be set directly through the `default_config` attribute. When set
    manually, it can be set to a dictionary or DictConfig object:

        >>> stage = Stage(map_fn=compute, default_config={"foo": "bar"})
        >>> print(f"{type(stage.default_config).__name__}: {stage.default_config}")
        DictConfig: {'foo': 'bar'}
        >>> stage = Stage(map_fn=compute, default_config=DictConfig({"foo_2": "bar_2"}))
        >>> print(f"{type(stage.default_config).__name__}: {stage.default_config}")
        DictConfig: {'foo_2': 'bar_2'}

    Or it can be set to a Path or str pointing to a YAML file containing the configuration:

        >>> config = DictConfig({"A": [1, 2, 3], "B": {"foo": "bar"}})
        >>> with tempfile.NamedTemporaryFile(suffix=".yaml") as tmpfile:
        ...     OmegaConf.save(config, tmpfile.name)
        ...     stage = Stage(map_fn=compute, default_config=tmpfile.name)
        >>> print(f"{type(stage.default_config).__name__}: {stage.default_config}")
        DictConfig: {'A': [1, 2, 3], 'B': {'foo': 'bar'}}

    If the corresponding file does not exist, an error will be raised:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     Stage(map_fn=compute, default_config=Path(tmpdir) / "foo.yaml")
        Traceback (most recent call last):
            ...
        FileNotFoundError: Default configuration file /tmp/tmp.../foo.yaml does not exist.

    Or if it is set to any other type, a TypeError will be raised:

        >>> Stage(map_fn=compute, default_config=42)
        Traceback (most recent call last):
            ...
        TypeError: Default configuration must be a dictionary, DictConfig, or path to a YAML file. Got
                   <class 'int'>: 42

    Proper set-up for test cases means the directory is or has any children that satisfy the
    `StageExample.is_example_dir` method.

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     example_dir = Path(tmpdir) / "examples"
        ...     example_dir.mkdir()
        ...     example_data = example_dir / "out_data.yaml"
        ...     _ = example_data.write_text("data/0.parquet: 'code,time,subject_id,numeric_value'")
        ...     stage = Stage(map_fn=compute, examples_dir=example_dir)
        ...     print(stage.test_cases)
        StageExample [base]
          stage_cfg: {}
          want_data:
            MEDSDataset:
            dataset_metadata:
            data_shards:
              - 0:
                pyarrow.Table
                subject_id: int64
                time: timestamp[us]
                code: string
                numeric_value: float
                ----
                subject_id: [[]]
                time: [[]]
                code: []
                numeric_value: [[]]
            code_metadata:
              pyarrow.Table
              code: string
              description: string
              parent_codes: list<item: string>
                child 0, item: string
              ----
              code: []
              description: []
              parent_codes: []
            subject_splits: None

    We can also have nested test cases:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     example_dir = Path(tmpdir) / "examples"
        ...     # Example # 1
        ...     ex_1 = example_dir / "example_1"
        ...     ex_1.mkdir(parents=True)
        ...     ex_1_data_fp = ex_1 / "out_data.yaml"
        ...     _ = ex_1_data_fp.write_text("data/0.parquet: 'code,time,subject_id,numeric_value'")
        ...     # Example # 2
        ...     ex_2 = example_dir / "example_2_foo"
        ...     ex_2.mkdir()
        ...     ex_2_metadata_fp = ex_2 / "out_metadata.yaml"
        ...     _ = ex_2_metadata_fp.write_text("metadata/codes.parquet: 'code,description,parent_codes'")
        ...     stage = Stage(map_fn=compute, examples_dir=example_dir)
        ...     print(stage.test_cases)
        example_2_foo:
        │   StageExample [base/example_2_foo]
        │     stage_cfg: {}
        │     want_metadata:
        │       shape: (0, 3)
        │       ┌──────┬─────────────┬──────────────┐
        │       │ code ┆ description ┆ parent_codes │
        │       │ ---  ┆ ---         ┆ ---          │
        │       │ str  ┆ str         ┆ list[str]    │
        │       ╞══════╪═════════════╪══════════════╡
        │       └──────┴─────────────┴──────────────┘
        example_1:
        │   StageExample [base/example_1]
        │     stage_cfg: {}
        │     want_data:
        │       MEDSDataset:
        │       dataset_metadata:
        │       data_shards:
        │         - 0:
        │           pyarrow.Table
        │           subject_id: int64
        │           time: timestamp[us]
        │          code: string
        │          numeric_value: float
        │          ----
        │          subject_id: [[]]
        │          time: [[]]
        │          code: []
        │          numeric_value: [[]]
        │       code_metadata:
        │         pyarrow.Table
        │         code: string
        │         description: string
        │         parent_codes: list<item: string>
        │           child 0, item: string
        │         ----
        │         code: []
        │         description: []
        │         parent_codes: []
        │       subject_splits: None

    The examples directory can also be inferred via a passed `_calling_file` argument, which is intended to be
    the file in which the `Stage.register` function was called. This looks to see if (1) the calling file
    lives in a parent directory of the same name as the defining stage and (2) if that directory has an
    `examples` subdirectory. If so, this subdirectory is et to the examples directory. Otherwise, nothing is
    set.

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     stage_dir = Path(tmpdir) / "stage_foo"
        ...     calling_file = stage_dir / "stage_foo.py"
        ...     examples_dir = stage_dir / "examples"
        ...     examples_dir.mkdir(parents=True)
        ...     calling_file.touch()
        ...     stage = Stage(map_fn=compute, stage_name="stage_foo", _calling_file=calling_file)
        ...     print(stage.examples_dir.relative_to(tmpdir))
        stage_foo/examples

    If the stage name doesn't match the directory name, we won't infer `self.examples_dir`:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     stage_dir = Path(tmpdir) / "stage_foo"
        ...     calling_file = stage_dir / "stage_foo.py"
        ...     examples_dir = stage_dir / "examples"
        ...     examples_dir.mkdir(parents=True)
        ...     calling_file.touch()
        ...     stage = Stage(map_fn=compute, stage_name="not_stage_foo", _calling_file=calling_file)
        ...     print(stage.examples_dir)
        None

    If the examples directory doesn't exist, we won't infer `self.examples_dir`:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     stage_dir = Path(tmpdir) / "stage_foo"
        ...     stage_dir.mkdir()
        ...     calling_file = stage_dir / "stage_foo.py"
        ...     calling_file.touch()
        ...     stage = Stage(map_fn=compute, stage_name="stage_foo", _calling_file=calling_file)
        ...     print(stage.examples_dir)
        None

    If the calling file doesn't exist, we won't infer `self.examples_dir`:

        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     stage_dir = Path(tmpdir) / "stage_foo"
        ...     calling_file = stage_dir / "stage_foo.py"
        ...     examples_dir = stage_dir / "examples"
        ...     examples_dir.mkdir(parents=True)
        ...     stage = Stage(map_fn=compute, stage_name="stage_foo", _calling_file=calling_file)
        ...     print(stage.examples_dir)
        None

    If the calling file is somehow misconfigured (e.g., it is a directory, not a file), a warning will be
    logged, and nothing will be inferred:

        >>> with tempfile.TemporaryDirectory() as tmpdir, print_warnings(), Stage.suppress_validation():
        ...     stage_dir = Path(tmpdir) / "stage_foo"
        ...     examples_dir = stage_dir / "examples"
        ...     examples_dir.mkdir(parents=True)
        ...     stage = Stage(map_fn=compute, stage_name="stage_foo", _calling_file=stage_dir)
        ...     print(stage.examples_dir)
        None
        Warning: Stage definition file /tmp/tmp.../stage_foo is not a file. Cannot infer examples
            directory.

    The stage, upon construction, attempts to validate that the stage is registered in the entry points. This
    can be disabled (as shown above) with the `Stage.suppress_validation` context manager, or by setting
    certain class variables. Had we disabled that, we would have seen

        >>> with print_warnings():
        ...     stage = Stage(map_fn=compute, stage_name="stage_foo")
        Warning: Stage 'stage_foo' is not registered in the entry points.
        This may be due to a missing or incorrectly configured entry point in your setup.py or pyproject.toml
        file. If this is during development, you may need to run `pip install -e .` to install your package
        properly in editable mode and ensure your stage registration is detected.
        You can disable this warning by setting the class variable `WARN_IF_NO_ENTRY_POINT_AT_NAME` to
        `False`, or filtering out `StageRegistrationWarning` warnings.
        You can disable all validation by setting the environment variable `DISABLE_STAGE_VALIDATION` to `1`.

    This warning indicates that there is no stage registered in the entry points with the name `stage_foo`. It
    is intended to flag to developers or users that they need to add their entry point to their package set-up
    to ensure things work correctly. We can also turn off this warning by directly setting the class variable
    `WARN_IF_NO_ENTRY_POINT_AT_NAME` to `False` on the `Stage` class or a derived class:

        >>> class StageNoWarn(Stage):
        ...     WARN_IF_NO_ENTRY_POINT_AT_NAME: ClassVar[bool] = False
        >>> with print_warnings():
        ...     stage = StageNoWarn(map_fn=compute, stage_name="stage_foo")

    Warnings aren't the end of the stage validation; if the stage is correctly configured, not only will
    `stage_foo` (or the stage name more generally) be registered, but it _will point to the Stage object being
    registered at this moment_. The best we can do for now to ensure this is to validate that if we try to
    load the Stage object being registered, we get an error due to attempting a circular import (as we would
    be attempting to reload the code that is running the stage construction). To show the limits of this,
    we'll patch out the stage registration object to return without errors and report a `stage_foo` being
    registered, and see what happens:

        >>> mock_stages = {"stage_foo": MagicMock()}
        >>> mock_stages["stage_foo"].load.side_effect = lambda: None
        >>> with patch("MEDS_transforms.stages.base.get_all_registered_stages", return_value=mock_stages):
        ...     with print_warnings():  # No warnings are printed as the stage is listed as being registered.
        ...         Stage(map_fn=compute, stage_name="stage_foo")
        Traceback (most recent call last):
            ...
        MEDS_transforms.stages.base.StageRegistrationError: Stage stage_foo is registered, but an attempted
        reload causes no issues. If this were the stage you are constructing, a reload would cause a circular
        import error, so this means you are overwriting an external, different stage, which is a problem!
        This may be due to a missing or incorrectly configured entry point in your setup.py or pyproject.toml
        file. If this is during development, you may need to run `pip install -e .` to install your package
        properly in editable mode and ensure your stage registration is detected.
        You can disable reload-should-cause-error checking by setting the class variable
        `ERR_IF_ENTRY_POINT_IMPORTABLE` to `False`.
        You can disable all validation by setting the environment variable `DISABLE_STAGE_VALIDATION` to `1`.

    These errors also occur if the loading of the stage raises an error that is not a circular import error.

        >>> def raise_unexpected_error():
        ...     raise AttributeError("unrelated")
        >>> mock_stages["stage_foo"].load.side_effect = raise_unexpected_error
        >>> with patch("MEDS_transforms.stages.base.get_all_registered_stages", return_value=mock_stages):
        ...     with print_warnings():  # No warnings are printed as the stage is listed as being registered.
        ...         Stage(map_fn=compute, stage_name="stage_foo")
        Traceback (most recent call last):
            ...
        ValueError: Failed to validate stage stage_foo for an unexpected reason; it is
        possible that a different stage is defined at this name.

    You can also turn off these errors by setting the class variable `ERR_IF_ENTRY_POINT_IMPORTABLE` to
    `False` on the `Stage` class or a derived class:

        >>> class StageNoErr(Stage):
        ...     ERR_IF_ENTRY_POINT_IMPORTABLE: ClassVar[bool] = False
        >>> with patch("MEDS_transforms.stages.base.get_all_registered_stages", return_value=mock_stages):
        ...     stage = StageNoErr(map_fn=compute, stage_name="stage_foo")
    """

    stage_type: StageType
    stage_name: str

    map_fn: ANY_COMPUTE_FN_T | None = None
    reduce_fn: ANY_COMPUTE_FN_T | None = None
    main_fn: MAIN_FN_T | None = None

    read_fn: READ_FN_T | None = None
    write_fn: WRITE_FN_T | None = None

    output_schema_updates: dict[str, pl.DataType] | None = None
    is_metadata: bool | None = None

    __mimic_fn: Callable | None = None
    __stage_docstring: str | None = None
    __stage_name: str | None = None
    __stage_dir: Path | None = None
    __examples_dir: Path | None = None
    __default_config: DictConfig | None = None

    WARN_IF_NO_ENTRY_POINT_AT_NAME: ClassVar[bool] = os.environ.get(VALIDATION_ENV_VAR, "0") != "1"
    ERR_IF_ENTRY_POINT_IMPORTABLE: ClassVar[bool] = os.environ.get(VALIDATION_ENV_VAR, "0") != "1"

    ENTRY_POINT_SETUP_STRING: ClassVar[str] = (
        "This may be due to a missing or incorrectly configured entry point in your setup.py or "
        "pyproject.toml file. If this is during development, you may need to run "
        "`pip install -e .` to install your package properly in editable mode and ensure your "
        "stage registration is detected. "
    )
    DISABLE_WARNING_STRING: ClassVar[str] = (
        "You can disable this warning by setting the class variable `WARN_IF_NO_ENTRY_POINT_AT_NAME` to "
        "`False`, or filtering out `StageRegistrationWarning` warnings. "
    )
    DISABLE_ERROR_STRING: ClassVar[str] = (
        "You can disable reload-should-cause-error checking by setting the class variable "
        "`ERR_IF_ENTRY_POINT_IMPORTABLE` to `False`."
    )
    DISABLE_ALL_STAGE_VALIDATION_STRING: ClassVar[str] = (
        "You can disable all validation by setting the environment variable "
        "`DISABLE_STAGE_VALIDATION` to `1`."
    )

    @staticmethod
    @contextmanager
    def suppress_validation():
        """Context manager to disable stage validation, to be used in testing and execution settings.

        Example:
            >>> with Stage.suppress_validation():
            ...     # Your code here that requires no stage validation.
            ...     pass
        """
        old_env_val = os.environ.get(VALIDATION_ENV_VAR, None)
        os.environ[VALIDATION_ENV_VAR] = "1"
        try:
            yield
        finally:
            if old_env_val is None:
                del os.environ[VALIDATION_ENV_VAR]
            else:
                os.environ[VALIDATION_ENV_VAR] = old_env_val

    def __init__(
        self,
        *,
        main_fn: MAIN_FN_T | None = None,
        map_fn: ANY_COMPUTE_FN_T | None = None,
        reduce_fn: ANY_COMPUTE_FN_T | None = None,
        read_fn: READ_FN_T | None = None,
        write_fn: WRITE_FN_T | None = None,
        stage_name: str | None = None,
        stage_docstring: str | None = None,
        output_schema_updates: dict[str, pl.DataType] | None = None,
        examples_dir: Path | None = None,
        default_config: dict[str, Any] | DictConfig | Path | str | None = None,
        is_metadata: bool | None = None,
        _calling_file: Path | None = None,
    ) -> MAIN_FN_T:
        """Wraps or returns a function that can serve as the main function for a stage."""

        self.stage_type = StageType.from_fns(main_fn, map_fn, reduce_fn)
        self.stage_name = stage_name
        self.stage_docstring = stage_docstring

        self.main_fn = main_fn
        self.map_fn = map_fn
        self.reduce_fn = reduce_fn

        self.read_fn = read_fn
        self.write_fn = write_fn

        if self.stage_type == StageType.MAIN and (self.read_fn is not None or self.write_fn is not None):
            raise ValueError(
                f"Stage type is {self.stage_type}, but read_fn or write_fn is set. This is only supported "
                "in MAP and MAPREDUCE stages. Please set read_fn and write_fn to None."
            )

        if is_metadata is None:
            if self.stage_type is StageType.MAPREDUCE:
                self.is_metadata = True
            elif self.stage_type is StageType.MAP:
                self.is_metadata = False
            else:
                raise ValueError(
                    "Stage type is not set to MAP or MAPREDUCE, but is_metadata is not set. Please set "
                    "is_metadata manually."
                )

            logger.debug(
                f"Automatically setting is_metadata={self.is_metadata} for {self.stage_name} as it is a "
                f"{self.stage_type} stage."
            )
        else:
            self.is_metadata = is_metadata

        self.__infer_stage_dir(_calling_file)

        self.examples_dir = examples_dir
        self.default_config = default_config

        do_skip_validation = (os.environ.get(VALIDATION_ENV_VAR, "0") == "1") or not (
            self.WARN_IF_NO_ENTRY_POINT_AT_NAME or self.ERR_IF_ENTRY_POINT_IMPORTABLE
        )

        if do_skip_validation:
            logger.debug(
                "Skipping stage validation at constructor time due to DISABLE_STAGE_VALIDATION environment "
                "variable being set. This is normal during execution of a stage via the MEDS-Transforms CLI, "
                "as validation happens manually in the main function in that context, but is typically not "
                "normal during testing, for example."
            )
        else:
            self.__validate_stage_entry_point_registration()

        if output_schema_updates is None:
            self.output_schema_updates = {}
        else:
            self.output_schema_updates = copy.deepcopy(output_schema_updates)

    def __infer_stage_dir(self, stage_definition_file: Path | None) -> Path | None:
        """Infers a possible stage directory based on the calling file.

        This is done by looking to see if the stage is defined in a file contained in a directory of the same
        name as the stage. If so, that directory is stored as the stage directory. If not, this will be None.
        """
        self.__stage_dir = None

        if stage_definition_file is None:
            return

        if not stage_definition_file.is_file():
            logger.warning(
                f"Stage definition file {stage_definition_file} is not a file. "
                "Cannot infer examples directory."
            )
            return

        # Get the directory of the calling file
        possible_stage_dir = stage_definition_file.parent

        if possible_stage_dir.name != self.stage_name:
            logger.debug(
                f"Stage definition file {stage_definition_file} is not in a directory with the same name as "
                "the stage {self.stage_name}. Cannot infer examples directory."
            )
            return

        self.__stage_dir = possible_stage_dir

    @property
    def examples_dir(self) -> Path | None:
        if self.__examples_dir is not None:
            return self.__examples_dir
        if self.__stage_dir is None:
            return None

        possible_examples_dir = self.__stage_dir / "examples"

        if possible_examples_dir.is_dir():
            return possible_examples_dir
        else:
            logger.debug(
                f"Stage definition file {self.__stage_dir} lacks an examples subdirectory. Can't infer "
                "examples directory."
            )
            return None

    @examples_dir.setter
    def examples_dir(self, examples_dir: Path | None):
        """Sets the examples directory for the stage.

        Args:
            examples_dir: The examples directory to set. This should be a directory containing nested
                test cases for the stage.
        """

        match examples_dir:
            case None | Path():
                self.__examples_dir = examples_dir
            case _:
                raise TypeError(
                    f"examples_dir must be a Path or None. Got {type(examples_dir)}: {examples_dir}"
                )

    @property
    def default_config(self) -> DictConfig:
        if self.__default_config is not None:
            return self.__default_config

        if self.__stage_dir is None:
            return DictConfig({})

        possible_default_config = self.__stage_dir / "config.yaml"
        if possible_default_config.is_file():
            return OmegaConf.load(possible_default_config)
        else:
            logger.debug(
                f"Stage definition file {self.__stage_dir} lacks a config.yaml file. Can't infer default "
                "configuration."
            )
            return DictConfig({})

    @default_config.setter
    def default_config(self, default_config: dict[str, Any] | DictConfig | Path | str | None):
        """Sets the default configuration for the stage.

        Args:
            default_config: The default configuration to set. This can be a dictionary, a DictConfig object,
                or a path to a YAML file containing the configuration.
        """
        match default_config:
            case None | DictConfig():
                self.__default_config = default_config
            case dict():
                self.__default_config = DictConfig(default_config)
            case Path() | str() as default_config_fp:
                if not Path(default_config_fp).is_file():
                    raise FileNotFoundError(f"Default configuration file {default_config_fp} does not exist.")
                self.__default_config = OmegaConf.load(default_config_fp)
            case _:
                raise TypeError(
                    "Default configuration must be a dictionary, DictConfig, or path to a YAML file. Got "
                    f"{type(default_config)}: {default_config}"
                )

    @property
    def test_cases(self) -> dict[str, StageExample]:
        if self.examples_dir is None:
            return {}

        examples_to_check = [self.examples_dir]
        test_cases = {}

        while examples_to_check:
            example_dir = examples_to_check.pop()

            if not example_dir.is_dir():
                continue

            if StageExample.is_example_dir(example_dir):
                scenario_name = example_dir.relative_to(self.examples_dir).as_posix()
                test_cases[scenario_name] = StageExample.from_dir(
                    stage_name=self.stage_name,
                    scenario_name=scenario_name,
                    example_dir=example_dir,
                    **self.output_schema_updates,
                )
            else:
                examples_to_check.extend(sorted(example_dir.iterdir()))

        return StageExampleDict(**test_cases)

    def __validate_stage_entry_point_registration(self):
        """Validates that the stage is registered in the entry points."""

        registered_stages = get_all_registered_stages()

        if self.stage_name not in registered_stages:
            if self.WARN_IF_NO_ENTRY_POINT_AT_NAME:
                # If the stage is not registered, we warn.
                logger.warning(
                    f"Stage '{self.stage_name}' is not registered in the entry points.\n"
                    f"{self.ENTRY_POINT_SETUP_STRING}\n{self.DISABLE_WARNING_STRING}\n"
                    f"{self.DISABLE_ALL_STAGE_VALIDATION_STRING}",
                )
            return

        if not self.ERR_IF_ENTRY_POINT_IMPORTABLE:
            return

        with Stage.suppress_validation():
            try:
                # Attempt to reload, which should cause an error if the stage being constructed is the same
                # stage as the one being registered.
                registered_stages[self.stage_name].load()
                raise StageRegistrationError(
                    f"Stage {self.stage_name} is registered, but an attempted reload causes "
                    "no issues. If this were the stage you are constructing, a reload would cause a circular "
                    "import error, so this means you are overwriting an external, different stage, which is "
                    "a problem!\n"
                    f"{self.ENTRY_POINT_SETUP_STRING}\n{self.DISABLE_ERROR_STRING}\n"
                    f"{self.DISABLE_ALL_STAGE_VALIDATION_STRING}"
                )
            except AttributeError as e:
                if "circular import" not in str(e):
                    raise ValueError(
                        f"Failed to validate stage {self.stage_name} for an unexpected reason; it is "
                        "possible that a different stage is defined at this name."
                    ) from e

    @property
    def stage_name(self) -> str:
        if self.__stage_name is not None:
            return self.__stage_name

        return (self.main_fn or self.map_fn).__module__.split(".")[-1]

    @stage_name.setter
    def stage_name(self, name: str):
        self.__stage_name = name

    @property
    def stage_docstring(self) -> str:
        """The docstring for the stage.

        This is set automatically based on the provided functions.
        """
        if self.__stage_docstring is not None:
            return self.__stage_docstring

        match self.stage_type:
            case StageType.MAIN:
                return self.main_fn_docstring
            case StageType.MAP:
                return self.map_fn_docstring
            case StageType.MAPREDUCE:
                return f"Map Stage:\n{self.map_fn_docstring}\n\nReduce stage:\n{self.reduce_fn_docstring}"

    @stage_docstring.setter
    def stage_docstring(self, docstring: str):
        self.__stage_docstring = docstring

    @property
    def main_fn_docstring(self) -> str:
        return (inspect.getdoc(self.main_fn) or "") if self.main_fn is not None else ""

    @property
    def map_fn_docstring(self) -> str:
        return (inspect.getdoc(self.map_fn) or "") if self.map_fn is not None else ""

    @property
    def reduce_fn_docstring(self) -> str:
        return (inspect.getdoc(self.reduce_fn) or "") if self.reduce_fn is not None else ""

    @property
    def main(self) -> MAIN_FN_T:
        match self.stage_type:
            case StageType.MAIN:

                @wraps(self.main_fn)
                def main_fn(cfg: DictConfig):
                    self.main_fn(cfg)

            case StageType.MAP:

                @wraps(self.map_fn)
                def main_fn(cfg: DictConfig):
                    map_stage(cfg, self.map_fn, read_fn=self.read_fn, write_fn=self.write_fn)

            case StageType.MAPREDUCE:

                def main_fn(cfg: DictConfig):
                    mapreduce_stage(
                        cfg, self.map_fn, self.reduce_fn, read_fn=self.read_fn, write_fn=self.write_fn
                    )

        main_fn.__name__ = self.stage_name
        main_fn.__doc__ = self.stage_docstring
        return main_fn

    @property
    def mimic_fn(self) -> Callable | None:
        """The function that this stage object should "mimic" when treated like a plain function.

        This is useful when a stage is constructed via decorator mode of `Stage.register`. In that case, we
        want this stage to mimic the decorated function upon normal usage. This practice enables doctests and
        imports of the stage function to work as expected. This property gets the function being mimicked.

        Examples:
            >>> def compute(cfg: DictConfig):
            ...     '''base compute docstring'''
            ...     return "compute"
            >>> def baz_fn(foo: str, bar: int):
            ...     '''base baz docstring'''
            ...     return f"baz {foo} {bar}"
            >>> stage = Stage.register(map_fn=compute)
            >>> print(stage.mimic_fn)
            None
            >>> stage.mimic_fn = baz_fn
            >>> stage.__name__
            'baz_fn'
            >>> stage.__doc__
            'base baz docstring'
            >>> stage("foo", 42)
            'baz foo 42'
            >>> stage.mimic_fn = "main"
            Traceback (most recent call last):
                ...
            TypeError: Cannot set mimic_fn to non-function. Got <class 'str'>
        """
        return self.__mimic_fn

    @mimic_fn.setter
    def mimic_fn(self, fn: Callable):
        """Sets the function that this stage object should "mimic" when treated like a plain function.

        In addition to the function itself, this also overwrites variables like `__name__` and `__doc__` to
        match the decorated function. This should really only be used to mimic one of the main or map
        functions, but this is not checked currently.

        Args:
            fn: The function to mimic.

        Raises:
            TypeError: If the provided argument is not a function.
        """

        if not inspect.isfunction(fn):
            raise TypeError(f"Cannot set mimic_fn to non-function. Got {type(fn)}")

        self.__mimic_fn = fn
        self.__name__ = fn.__name__
        self.__doc__ = fn.__doc__

        # Fix doctests
        try:
            module = sys.modules[fn.__module__]
            if hasattr(module, "__test__"):  # pragma: no cover
                module.__test__[fn.__name__] = fn
            else:
                module.__test__ = {fn.__name__: fn}
        except Exception as e:  # pragma: no cover
            warnings.warn(f"Failed to set doctest for {fn.__name__}: {e}", stacklevel=2)

    def __call__(self, *args, **kwargs):
        if self.mimic_fn is not None:
            return self.mimic_fn(*args, **kwargs)

        if len(args) == 1 and len(kwargs) == 0 and inspect.isfunction(args[0]):
            # This is likely inappropriate decorator usage, so we raise a custom error.
            raise ValueError(
                "If a Stage is constructed via keyword arguments directly (including a `main_fn` or a "
                "`map_fn`), then it cannot be used as a decorator or otherwise called directly."
            )

        raise ValueError(f"Stage {self.stage_name} has no function to mimic, so can't be called directly.")

    def __str__(self) -> str:
        lines = [
            f"Stage {self.stage_name}:",
            f"  Type: {self.stage_type}",
            f"  is_metadata: {self.is_metadata}",
            "  Docstring:",
        ]

        pretty_wrap = partial(textwrap.wrap, width=110, initial_indent="    | ", subsequent_indent="    | ")

        docstring_lines = textwrap.dedent(self.stage_docstring).splitlines()
        for line in docstring_lines:
            lines.extend(pretty_wrap(line))

        if self.default_config:
            lines.append("  Default config:")
            lines.extend(textwrap.indent(str(OmegaConf.to_yaml(self.default_config)), "    | ").splitlines())

        if self.output_schema_updates:
            lines.append("  Output schema updates:")
            lines.extend(pretty_wrap(str(self.output_schema_updates)))

        lines.extend(
            [
                f"  Map function: {self.map_fn.__name__ if self.map_fn else None}",
                f"  Reduce function: {self.reduce_fn.__name__ if self.reduce_fn else None}",
                f"  Main function: {self.main_fn.__name__ if self.main_fn else None}",
                f"  Mimic function: {self.mimic_fn.__name__ if self.mimic_fn else None}",
            ]
        )
        return "\n".join(lines)

    @classmethod
    def register(cls, *args, **kwargs) -> Callable[[Callable], Stage] | Stage:
        """This method used to define and register a MEDS-Transforms stage of any variety.

        ## Function usage modes:

        It can be used either as a decorator, a parametrized decorator, or a direct method, depending on the
        arguments provided and manner of invocation.

        ### As a decorator
        If used as a decorator (e.g., `@Stage.register`) on a single function, it will define a stage using
        that function. The stage defined will either be a `StageType.MAIN` stage if the function is named
        `main`, or a `StageType.MAP` stage otherwise. The return value will be a Stage object set to mimic the
        called function.

        ### As a parametrized decorator
        If used as a parametrized decorator (e.g., `@Stage.register(stage_name="foo")`), it will define a
        stage using the decorated function, but with the specified keyword arguments modifying the stage
        appropriately. The keyword arguments included can include any property save `main_fn` or `map_fn`. The
        stage will either be a `StageType.MAIN` stage if the function is named `main` (in which case a
        `reduce_fn` property cannot be set), a `StageType.MAP` stage if the decorated function is not named
        main and a `reduce_fn` is not set, or a `StageType.MAPREDUCE` stage if the decorated function is not
        named `main` and a `reduce_fn` is set. The return value will be a Stage object set to mimic the
        decorated function. If a `reduce_fn` is passed as a keyword argument, it will not be modified.

        ### As a direct method
        If used as a direct method, (e.g., `main = Stage.register(main_fn=main)`), it will define a stage
        using the provided arguments, and return the stage object. The returned object will not be set to
        mimic any function, and the `main` method will be the main entry point for the stage. This mode is
        only activated if sufficient keyword arguments are provided to define a stage (meaning a `main_fn` or
        `map_fn` must be set). The passed functions will not be modified.

        Args:
            *args: Positional arguments. If used as a decorator, this should be a single function, and no
                keyword arguments should be set.
            **kwargs: Keyword arguments. These can include all keyword arguments to the `Stage` constructor
                save for `_calling_file`; namely, `main_fn`, `map_fn`, `reduce_fn`, `stage_name`,
                `stage_docstring`, `examples_dir`, `output_schema_updates`, and `default_config`.
                Not all keyword arguments are required for all usages of the decorator.

        ## Inference of static data examples and the default configuration filepath.

        When this function is called, it will attempt to locate the filepath of the file in which it was
        called, and pass that as the `_calling_file` argument to the `Stage` constructor. This is used to
        infer the examples directory and default configuration option if thery are not manually set. Note that
        this means that the parameter `_calling_file` can not be used as a keyword argument to this function.
        You should never need this parameter anyways, as if you are setting something manually you can
        directly set the example directory.

        Returns:
            Either a `Stage` object or a decorator function, depending on the arguments provided.

        Raises:
            ValueError: If multiple positional arguments are passed or if the function is not callable.
            TypeError: If the first positional argument is not a function.
            ValueError: If both `main_fn` and `map_fn` or `reduce_fn` are provided.

        Examples:
            >>> def main(cfg: DictConfig):
            ...     '''base main docstring'''
            ...     return "main"
            >>> def map_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base map docstring'''
            ...     return "map"
            >>> def reduce_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base reduce docstring'''
            ...     return "reduce"
            >>> stage = Stage.register(main_fn=main, is_metadata=True)
            >>> print(stage)  # The name is inferred from the name of the file:
            Stage base:
              Type: main
              is_metadata: True
              Docstring:
                | base main docstring
              Map function: None
              Reduce function: None
              Main function: main
              Mimic function: None


        This shows direct method usage: When used with only keyword arguments that fully define a function a
        stage set to mimic nothing is returned directly, with the parameters set as defined by what kind of
        stage is being created. As it has no mimic function, if you try to call it directly, it will raise an
        error, and not act like the main function.

            >>> main({})
            'main'
            >>> stage({})
            Traceback (most recent call last):
                ...
            ValueError: Stage base has no function to mimic, so can't be called directly.

        This logic applies regardless of what kind of stage you register; e.g., a map or mapreduce stage will
        likewise have no mimic function when defined in this way:

            >>> print(Stage.register(map_fn=map_fn, is_metadata=False))
            Stage base:
              Type: map
              is_metadata: False
              Docstring:
                | base map docstring
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: None
            >>> print(Stage.register(map_fn=map_fn, reduce_fn=reduce_fn, is_metadata=True))
            Stage base:
              Type: mapreduce
              is_metadata: True
              Docstring:
                | Map Stage:
                | base map docstring
                | Reduce stage:
                | base reduce docstring
              Map function: map_fn
              Reduce function: reduce_fn
              Main function: None
              Mimic function: None

        If you call it with main and map functions or a main and a reduce function, it will raise an error:

            >>> Stage.register(main_fn=main, map_fn=map_fn)
            Traceback (most recent call last):
                ...
            ValueError: Only one of main_fn or map_fn/reduce_fn should be provided.
            >>> Stage.register(main_fn=main, reduce_fn=reduce_fn)
            Traceback (most recent call last):
                ...
            ValueError: Only one of main_fn or map_fn/reduce_fn should be provided.

        The `is_metadata` parameter is a bit special, in that you need to provide it if you are defining a
        "main" stage; otherwise it can be omitted and inferred to be `True` if the stage is a mapreduce stage,
        or `False` if it is a map stage.

            >>> print(Stage.register(map_fn=map_fn))
            Stage base:
              Type: map
              is_metadata: False
              Docstring:
                | base map docstring
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: None
            >>> print(Stage.register(map_fn=map_fn, reduce_fn=reduce_fn))
            Stage base:
              Type: mapreduce
              is_metadata: True
              Docstring:
                | Map Stage:
                | base map docstring
                | Reduce stage:
                | base reduce docstring
              Map function: map_fn
              Reduce function: reduce_fn
              Main function: None
              Mimic function: None
            >>> Stage.register(main_fn=main)
            Traceback (most recent call last):
                ...
            ValueError: Stage type is not set to MAP or MAPREDUCE, but is_metadata is not set. Please set
            is_metadata manually.

        Another set of special parameters are the `read_fn` and `write_fn` parameters. These can be used to
        specialize the behavior of map and mapreduce stages, but cannot be set during on main stages:

            >>> def write_fn(df: pl.DataFrame, fp: Path):
            ...     df.write_parquet(fp)
            >>> stage = Stage.register(map_fn=map_fn, write_fn=write_fn)
            >>> print(stage.write_fn)
            <function write_fn at ...>
            >>> print(stage)
            Stage base:
              Type: map
              is_metadata: False
              Docstring:
                | base map docstring
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: None
            >>> Stage.register(main_fn=main, write_fn=write_fn)
            Traceback (most recent call last):
                ...
            ValueError: Stage type is main, but read_fn or write_fn is set. This is only supported in MAP
                and MAPREDUCE stages. Please set read_fn and write_fn to None.

        You can also use `Stage.register` as a decorator, either parametrized or not.

            >>> @Stage.register
            ... def map_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base map docstring'''
            ...     return "map"
            >>> @Stage.register(is_metadata=True)
            ... def main(cfg: DictConfig):
            ...     '''base main docstring'''
            ...     return "main"

        The output of the decorator, which is stored in the variable of the name defined by the function, will
        "mimic" the decorated function under normal usage:

            >>> map_fn.__name__
            'map_fn'
            >>> map_fn.__doc__
            'base map docstring'
            >>> map_fn({}, {})
            'map'
            >>> main.__name__
            'main'
            >>> main.__doc__
            'base main docstring'
            >>> main({})
            'main'

        ... but it is actually a stage object defined by the decorator:

            >>> print(map_fn)
            Stage base:
              Type: map
              is_metadata: False
              Docstring:
                | base map docstring
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: map_fn
            >>> print(main)
            Stage base:
              Type: main
              is_metadata: True
              Docstring:
                | base main docstring
              Map function: None
              Reduce function: None
              Main function: main
              Mimic function: main

        When used as a decorator, you can also parametrize the decorator with other parameters:

            >>> @Stage.register(stage_name="foo", stage_docstring="bar", is_metadata=False)
            ... def main(cfg: DictConfig):
            ...     '''base main docstring'''
            ...     return "main"
            >>> print(main)
            Stage foo:
              Type: main
              is_metadata: False
              Docstring:
                | bar
              Map function: None
              Reduce function: None
              Main function: main
              Mimic function: main

        The acceptable keyword arguments to the decorator are the same as those to the constructor, except for
        `_calling_file` (which is automatically inferred and will be discussed more later). For example:

            >>> @Stage.register(
            ...     stage_name="foo",
            ...     stage_docstring="bar",
            ...     output_schema_updates={"foo": pl.Int64},
            ...     examples_dir=Path("foo"),
            ...     default_config={"arg1": {"option1": "foo"}, "arg2": [1, 2.3]},
            ... )
            ... def foobar(cfg: DictConfig):
            ...     '''base map docstring'''
            ...     return "map"
            >>> print(foobar)
            Stage foo:
              Type: map
              is_metadata: False
              Docstring:
                | bar
              Default config:
                | arg1:
                |   option1: foo
                | arg2:
                | - 1
                | - 2.3
              Output schema updates:
                | {'foo': Int64}
              Map function: foobar
              Reduce function: None
              Main function: None
              Mimic function: foobar

        The decorated function will be inferred to be a "main" function if it is named "main" and there is no
        reduce function specified in the parametrized keyword arguments to the decorator. Otherwise, it will
        be assumed to be a map function:

            >>> @Stage.register
            ... def map_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base map docstring'''
            ...     return "map"
            >>> print(map_fn)
            Stage base:
              Type: map
              is_metadata: False
              Docstring:
                | base map docstring
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: map_fn
            >>> @Stage.register(reduce_fn=reduce_fn)
            ... def main(cfg: DictConfig):
            ...     '''base main docstring'''
            ...     return "main"
            >>> print(main)
            Stage base:
              Type: mapreduce
              is_metadata: True
              Docstring:
                | Map Stage:
                | base main docstring
                | Reduce stage:
                | base reduce docstring
              Map function: main
              Reduce function: reduce_fn
              Main function: None
              Mimic function: main

        You can't use it as a decorator while specifying either the main or map function in the keyword
        arguments:

            >>> @Stage.register(map_fn=map_fn)
            ... def reduce_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base reduce docstring'''
            ...     return "reduce"
            Traceback (most recent call last):
                ...
            ValueError: If a Stage is constructed via keyword arguments directly (including a `main_fn` or a
            `map_fn`), then it cannot be used as a decorator or otherwise called directly.

        You also can't specify a positional argument and keyword arguments at the same time or multiple
        positional arguments:

            >>> stage = Stage.register(main, stage_name="foo")
            Traceback (most recent call last):
                ...
            ValueError: Cannot provide keyword arguments when using as a decorator.
            >>> stage = Stage.register(main, map_fn)
            Traceback (most recent call last):
                ...
            ValueError: Stage.register can only be used with at most a single positional arg. Got 2

        Though not recommended, you could theoretically use decorator mode directly to create a function that
        would return a stage. This exposes some further error surfaces which are checked explicitly:

            >>> def main(cfg: DictConfig):
            ...     '''base main docstring'''
            ...     return "main"
            >>> def map_fn(cfg: DictConfig, stage_cfg: DictConfig):
            ...     '''base map docstring'''
            ...     return "map"
            >>> manual_decorator = Stage.register(stage_name="foo", stage_docstring="bar")
            >>> print(manual_decorator(map_fn))
            Stage foo:
              Type: map
              is_metadata: False
              Docstring:
                | bar
              Map function: map_fn
              Reduce function: None
              Main function: None
              Mimic function: map_fn
            >>> manual_decorator(main, map_fn=map_fn)
            Traceback (most recent call last):
                ...
            ValueError: Cannot provide main_fn or map_fn kwargs when using as a decorator.
            >>> manual_decorator("foo")
            Traceback (most recent call last):
                ...
            TypeError: First argument must be a function. Got <class 'str'>

        The parameter `_calling_file` is used to infer the examples directory, and can't be set manually in
        the function.

            >>> Stage.register(_calling_file="foo")
            Traceback (most recent call last):
                ...
            ValueError: Cannot provide keyword arguments that are also automatically inferred. Got
            {'_calling_file'}

        This is because this function sets the calling file to the location in the code where the decorator is
        called, and from that the stage will try to infer the "stage directory" and the associated default
        configuration file path and examples directory, if possible. We can demonstrate how this works by
        patching out the inspect module to return a file path that we'll create here:

            >>> config = DictConfig({"arg1": {"option1": "foo"}, "arg2": [1, 2.3, None], "arg3": None})
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            ...     # Make the stage directory:
            ...     stage_dir = Path(tmpdir) / "stage_foo"
            ...     stage_dir.mkdir()
            ...     # Make the calling file:
            ...     calling_file = stage_dir / "stage_foo.py"
            ...     calling_file.touch()
            ...     # Make the examples directory:
            ...     examples_dir = stage_dir / "examples"
            ...     # We'll make two example cases:
            ...     example_1_dir = examples_dir / "example_1"
            ...     example_1_dir.mkdir(parents=True)
            ...     example_1_fp = example_1_dir / "out_data.yaml"
            ...     _ = example_1_fp.write_text("data/0.parquet: 'code,time,subject_id,numeric_value'")
            ...     example_2_dir = examples_dir / "example_2"
            ...     example_2_dir.mkdir(parents=True)
            ...     example_2_fp = example_2_dir / "out_data.yaml"
            ...     _ = example_2_fp.write_text("data/1.parquet: 'code,time,subject_id,numeric_value'")
            ...     # Save the config file:
            ...     default_config_file = stage_dir / "config.yaml"
            ...     OmegaConf.save(config, default_config_file)
            ...     # What does the directory structure look like?
            ...     print("Root directory:")
            ...     print("-----------------")
            ...     print_directory(tmpdir)
            ...     # Mock out the inspect module to return the calling file we're constructing:
            ...     with patch("inspect.currentframe") as mock:
            ...         mock.return_value.f_back.f_code.co_filename = str(calling_file)
            ...
            ...         # Now we can create the stage:
            ...         @Stage.register(stage_name="stage_foo", is_metadata=True)
            ...         def main(cfg: DictConfig):
            ...             '''base main docstring'''
            ...             return "main"
            ...
            ...         # Print the stage object and see if it has set the examples directory and default
            ...         print("-----------------")
            ...         print("Stage object:")
            ...         print("-----------------")
            ...         print(main)
            ...         # Check the test cases, which aren't loaded with the stage by default as they aren't
            ...         # necessary during normal runtime:
            ...         print("-----------------")
            ...         print("Test cases:")
            ...         print("-----------------")
            ...         print(main.test_cases)
            Root directory:
            -----------------
            └── stage_foo
                ├── config.yaml
                ├── examples
                │   ├── example_1
                │   │   └── out_data.yaml
                │   └── example_2
                │       └── out_data.yaml
                └── stage_foo.py
            -----------------
            Stage object:
            -----------------
            Stage stage_foo:
              Type: main
              is_metadata: True
              Docstring:
                | base main docstring
              Default config:
                | arg1:
                |   option1: foo
                | arg2:
                | - 1
                | - 2.3
                | - null
                | arg3: null
              Map function: None
              Reduce function: None
              Main function: main
              Mimic function: main
            -----------------
            Test cases:
            -----------------
            example_2:
            │   StageExample [stage_foo/example_2]
            │     stage_cfg: {}
            │     want_data:
            │       MEDSDataset:
            │       dataset_metadata:
            │       data_shards:
            │         - 1:
            │           pyarrow.Table
            │           subject_id: int64
            │           time: timestamp[us]
            │           code: string
            │           numeric_value: float
            │           ----
            │           subject_id: [[]]
            │           time: [[]]
            │           code: []
            │           numeric_value: [[]]
            │       code_metadata:
            │         pyarrow.Table
            │         code: string
            │         description: string
            │         parent_codes: list<item: string>
            │           child 0, item: string
            │         ----
            │         code: []
            │         description: []
            │         parent_codes: []
            │       subject_splits: None
            example_1:
            │   StageExample [stage_foo/example_1]
            │     stage_cfg: {}
            │     want_data:
            │       MEDSDataset:
            │       dataset_metadata:
            │       data_shards:
            │         - 0:
            │           pyarrow.Table
            │           subject_id: int64
            │           time: timestamp[us]
            │           code: string
            │           numeric_value: float
            │           ----
            │           subject_id: [[]]
            │           time: [[]]
            │           code: []
            │           numeric_value: [[]]
            │       code_metadata:
            │         pyarrow.Table
            │         code: string
            │         description: string
            │         parent_codes: list<item: string>
            │           child 0, item: string
            │         ----
            │         code: []
            │         description: []
            │         parent_codes: []
            │       subject_splits: None

        If the example set-up is the same, but the stage name doesn't agree with the parent directory of the
        calling file, then we get a normal stage with none of the extra information as the examples directory
        and config.yaml file are assumed to be unrelated:

            >>> config = DictConfig({"arg1": {"option1": "foo"}, "arg2": [1, 2.3, None], "arg3": None})
            >>> with tempfile.TemporaryDirectory() as tmpdir:
            ...     # Make the stage directory:
            ...     stage_dir = Path(tmpdir) / "stage_foo"
            ...     stage_dir.mkdir()
            ...     # Make the calling file:
            ...     calling_file = stage_dir / "stage_foo.py"
            ...     calling_file.touch()
            ...     # Make the examples directory:
            ...     examples_dir = stage_dir / "examples"
            ...     # We'll make two example cases:
            ...     example_1_dir = examples_dir / "example_1"
            ...     example_1_dir.mkdir(parents=True)
            ...     example_1_fp = example_1_dir / "out_data.yaml"
            ...     _ = example_1_fp.write_text("data/0.parquet: 'code,time,subject_id,numeric_value'")
            ...     example_2_dir = examples_dir / "example_2"
            ...     example_2_dir.mkdir(parents=True)
            ...     example_2_fp = example_2_dir / "out_data.yaml"
            ...     _ = example_2_fp.write_text("data/1.parquet: 'code,time,subject_id,numeric_value'")
            ...     # Save the config file:
            ...     default_config_file = stage_dir / "config.yaml"
            ...     OmegaConf.save(config, default_config_file)
            ...     # What does the directory structure look like?
            ...     print("Root directory:")
            ...     print("-----------------")
            ...     print_directory(tmpdir)
            ...     # Mock out the inspect module to return the calling file we're constructing:
            ...     with patch("inspect.currentframe") as mock:
            ...         mock.return_value.f_back.f_code.co_filename = str(calling_file)
            ...
            ...         # Now we can create the stage:
            ...         @Stage.register(stage_name="not_stage_foo", is_metadata=True)
            ...         def main(cfg: DictConfig):
            ...             '''base main docstring'''
            ...             return "main"
            ...
            ...         # Print the stage object and see if it has set the examples directory and default
            ...         print("-----------------")
            ...         print("Stage object:")
            ...         print("-----------------")
            ...         print(main)
            ...         # Check the test cases, which aren't loaded with the stage by default as they aren't
            ...         # necessary during normal runtime:
            ...         print("-----------------")
            ...         print("Test cases:")
            ...         print("-----------------")
            ...         print(main.test_cases)
            Root directory:
            -----------------
            └── stage_foo
                ├── config.yaml
                ├── examples
                │   ├── example_1
                │   │   └── out_data.yaml
                │   └── example_2
                │       └── out_data.yaml
                └── stage_foo.py
            -----------------
            Stage object:
            -----------------
            Stage not_stage_foo:
              Type: main
              is_metadata: True
              Docstring:
                | base main docstring
              Map function: None
              Reduce function: None
              Main function: main
              Mimic function: main
            -----------------
            Test cases:
            -----------------
            {}

        What about those warnings? To see the warnings in this test, we'll explicitly capture and print logged
        warnings. These will normally only be visible in the logger output, but we'll capture and print them
        here using the `print_warnings` context manager, defined in our `conftest.py` file.

            >>> with print_warnings():
            ...
            ...     @Stage.register(is_metadata=True)
            ...     def main(cfg: DictConfig):
            ...         '''base main docstring'''
            ...         return "main"
            Warning: Stage 'base' is not registered in the entry points. This may be due to a missing or
            incorrectly configured entry point in your setup.py or pyproject.toml file. If this is during
            development, you may need to run `pip install -e .` to install your package properly in editable
            mode and ensure your stage registration is detected. You can disable this warning by setting the
            class variable `WARN_IF_NO_ENTRY_POINT_AT_NAME` to `False`, or filtering out
            `StageRegistrationWarning` warnings. You can disable all validation by setting the environment
            variable `DISABLE_STAGE_VALIDATION` to `1`.
        """

        # Get the frame of the caller
        caller_frame = inspect.currentframe().f_back
        # Get the filename from the frame
        calling_file = Path(caller_frame.f_code.co_filename)
        if not calling_file.is_file():  # pragma: no cover
            # In this case, something is wrong and we can't infer what the file is, so we omit it. This mostly
            # comes up in test cases, not real usage.
            calling_file = None

        inferred_kwargs = {"_calling_file": calling_file}

        if set(inferred_kwargs.keys()).intersection(kwargs.keys()):
            raise ValueError(
                "Cannot provide keyword arguments that are also automatically inferred. "
                f"Got {set(inferred_kwargs).intersection(kwargs.keys())}"
            )

        def decorator(fn: Callable, **kwargs):
            if not inspect.isfunction(fn):
                raise TypeError(f"First argument must be a function. Got {type(fn)}")

            if "main_fn" in kwargs or "map_fn" in kwargs:
                raise ValueError("Cannot provide main_fn or map_fn kwargs when using as a decorator.")

            stage_kwargs = {**kwargs}

            if (fn.__name__ == "main") and ("reduce_fn" not in kwargs):
                stage_kwargs["main_fn"] = fn
            else:
                stage_kwargs["map_fn"] = fn

            stage = Stage(**stage_kwargs)
            stage.mimic_fn = fn
            return stage

        if len(args) == 0:
            if "main_fn" in kwargs or "map_fn" in kwargs:
                return Stage(**kwargs, **inferred_kwargs)
            return partial(decorator, **kwargs, **inferred_kwargs)
        elif len(args) == 1:
            if kwargs:
                raise ValueError("Cannot provide keyword arguments when using as a decorator.")
            return decorator(args[0], **inferred_kwargs)
        else:
            raise ValueError(
                f"Stage.register can only be used with at most a single positional arg. Got {len(args)}"
            )



"""This file defines the structured base classes for the various configs used in MEDS-Transforms."""

from __future__ import annotations

import dataclasses
import json
import logging
from pathlib import Path

from meds import DatasetMetadataSchema, dataset_metadata_filepath

from .utils import OmegaConfResolver, hydra_registered_dataclass

logger = logging.getLogger(__name__)


def get_dataset_metadata_from_root(root: str) -> DatasetMetadataSchema:
    """Get the dataset metadata from the MEDS root directory.

    Args:
        root (str): The root directory of the dataset.

    Returns:
        dict[str, str]: The dataset metadata.

    Raises:
        FileNotFoundError: If the dataset metadata file is not found.

    Examples:
        >>> metadata = DatasetMetadataSchema(
        ...     dataset_name="example_dataset",
        ...     dataset_version="1.0.0",
        ... )
        >>> with tempfile.TemporaryDirectory() as MEDS_root:
        ...     metadata_fp = Path(MEDS_root) / dataset_metadata_filepath
        ...     metadata_fp.parent.mkdir(parents=True, exist_ok=False)
        ...     _ = metadata_fp.write_text(json.dumps(metadata.to_dict()))
        ...     get_dataset_metadata_from_root(MEDS_root)
        {'dataset_name': 'example_dataset', 'dataset_version': '1.0.0'}

        Errors will be raised if the dataset metadata file is not found:

        >>> with tempfile.TemporaryDirectory() as MEDS_root:
        ...     get_dataset_metadata_from_root(MEDS_root)
        Traceback (most recent call last):
            ...
        FileNotFoundError: Dataset metadata file not found at /tmp/tmp.../metadata/dataset.json
    """
    fp = Path(root) / dataset_metadata_filepath
    if not fp.exists():
        raise FileNotFoundError(f"Dataset metadata file not found at {fp}")
    return DatasetMetadataSchema(**json.loads(fp.read_text())).to_dict()


@OmegaConfResolver
def get_dataset_name_from_root(root: str, default: str = "Unknown") -> str:
    """Get the dataset name from the root directory.

    Args:
        root (str): The root directory of the dataset.
        default (str): The default value to return if the dataset name is not found.

    Returns:
        str: The dataset name, or the default value if not found.

    Raises:
        DatasetMetadataNotFoundWarning: If the dataset metadata file is not found or is invalid.

    Examples:
        >>> metadata = DatasetMetadataSchema(
        ...     dataset_name="example_dataset",
        ...     dataset_version="1.0.0",
        ... )
        >>> with tempfile.TemporaryDirectory() as MEDS_root:
        ...     metadata_fp = Path(MEDS_root) / dataset_metadata_filepath
        ...     metadata_fp.parent.mkdir(parents=True, exist_ok=False)
        ...     _ = metadata_fp.write_text(json.dumps(metadata.to_dict()))
        ...     get_dataset_name_from_root(MEDS_root)
        'example_dataset'

        If the dataset metadata file is not found or can't be parsed, the default is returned and a warning is
        logged, which we can catch and print with the `print_warnings` context manager (defined in our
        `conftest.py`):

        >>> metadata = DatasetMetadataSchema(
        ...     dataset_name="example_dataset",
        ...     dataset_version="1.0.0",
        ... )
        >>> with print_warnings(), tempfile.TemporaryDirectory() as MEDS_root:
        ...     get_dataset_name_from_root(MEDS_root, default="Foo")
        'Foo'
        Warning: Valid dataset metadata file not found in /tmp/tmp...:
                 Dataset metadata file not found at /tmp/tmp.../metadata/dataset.json
        >>> with print_warnings(), tempfile.TemporaryDirectory() as MEDS_root:
        ...     metadata_fp = Path(MEDS_root) / dataset_metadata_filepath
        ...     metadata_fp.parent.mkdir(parents=True, exist_ok=False)
        ...     _ = metadata_fp.write_text("def foo(): return 42")  # Invalid JSON
        ...     get_dataset_name_from_root(MEDS_root, default="Foo")
        'Foo'
        Warning: Valid dataset metadata file not found in /tmp/tmp...:
                 Expecting value: line 1 column 1 (char 0)
    """
    try:
        return get_dataset_metadata_from_root(root).get("dataset_name", default)
    except Exception as e:
        logger.warning(f"Valid dataset metadata file not found in {root}: {e}")
        return default


@OmegaConfResolver
def get_dataset_version_from_root(root: str, default: str = "Unknown") -> str:
    """Get the dataset version from the root directory.

    Args:
        root (str): The root directory of the dataset.
        default (str): The default value to return if the dataset name is not found.

    Returns:
        str: The dataset version, or the default value if not found.

    Raises:
        DatasetMetadataNotFoundWarning: If the dataset metadata file is not found or is invalid.

    Examples:
        >>> metadata = DatasetMetadataSchema(
        ...     dataset_name="example_dataset",
        ...     dataset_version="1.0.0",
        ... )
        >>> with tempfile.TemporaryDirectory() as MEDS_root:
        ...     metadata_fp = Path(MEDS_root) / dataset_metadata_filepath
        ...     metadata_fp.parent.mkdir(parents=True, exist_ok=False)
        ...     _ = metadata_fp.write_text(json.dumps(metadata.to_dict()))
        ...     get_dataset_version_from_root(MEDS_root)
        '1.0.0'

        If the dataset metadata file is not found or can't be parsed, the default is returned and a warning is
        logged, which we can catch and print with the `print_warnings` context manager (defined in our
        `conftest.py`):

        >>> metadata = DatasetMetadataSchema(
        ...     dataset_name="example_dataset",
        ...     dataset_version="1.0.0",
        ... )
        >>> with print_warnings(), tempfile.TemporaryDirectory() as MEDS_root:
        ...     metadata_fp = Path(MEDS_root) / dataset_metadata_filepath
        ...     metadata_fp.parent.mkdir(parents=True, exist_ok=False)
        ...     _ = metadata_fp.write_text("Hello world!")  # Invalid JSON
        ...     get_dataset_version_from_root(MEDS_root, default="Bar")
        'Bar'
        Warning: Valid dataset metadata file not found in /tmp/tmp...:
                 Expecting value: line 1 column 1 (char 0)
        >>> with print_warnings(), tempfile.TemporaryDirectory() as MEDS_root:
        ...     get_dataset_version_from_root(MEDS_root, default="Bar")
        'Bar'
        Warning: Valid dataset metadata file not found in /tmp/tmp...:
                 Dataset metadata file not found at /tmp/tmp.../metadata/dataset.json
    """
    try:
        return get_dataset_metadata_from_root(root).get("dataset_version", default)
    except Exception as e:
        logger.warning(f"Valid dataset metadata file not found in {root}: {e}")
        return default


@hydra_registered_dataclass(group="dataset", name="_base_dataset")
class DatasetConfig:
    """A base configuration class for MEDS dataset inputs.

    This class is used to define the base configuration for a dataset (largely for type safety purposes). It
    includes the root directory of the dataset, and the name and version of the dataset. This is merely a base
    class used for type safety in the Hydra configs. In hydra configuration usage, the resolvers defined below
    populate the name and version parameters given the root dir automatically, if possible.

    Attributes:
        root_dir: The root directory of the dataset.
        name: The name of the dataset.
        version: The version of the dataset.
        code_modifiers: A list of code modifiers for use when processing to the dataset.
    """

    root_dir: str
    name: str = "${oc:get_dataset_name_from_root:root_dir}"
    version: str = "${oc:get_dataset_version_from_root:root_dir}"
    code_modifiers: list[str] = dataclasses.field(default_factory=list)



from .compute_fn import ANY_COMPUTE_FN_T, COMPUTE_FN_T, bind_compute_fn
from .match_revise import is_match_revise, match_revise_fntr

__all__ = ["bind_compute_fn", "is_match_revise", "match_revise_fntr"]
